""" Unit Tests for Py-ART's graph/plot_cfradial.py module. """
# execute this script to create figure_plot_cfradial_*.png files.
# To compare results
# display figure_plot_netcdf_ppi.png&
# display figure_plot_cfradial_ppi_radar.png&
# display figure_plot_netcdf_rhi.png&
# display figure_plot_cfradial_rhi_radar.png&

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import netCDF4
import matplotlib.pyplot as plt
import pyart


def test_plot_netcdf_rhi(outfile=None):
    dataset = netCDF4.Dataset(pyart.testing.CFRADIAL_RHI_FILE)
    display = pyart.graph.CFRadialDisplay(dataset)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('reflectivity_horizontal', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_netcdf_ppi(outfile=None):
    dataset = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.CFRadialDisplay(dataset)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity_horizontal', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_netcdf_ppi_masked(outfile=None):
    dataset = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.CFRadialDisplay(dataset)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity_horizontal', 0, ax=ax,
                     mask_tuple=('reflectivity_horizontal', 16.0))
    if outfile:
        fig.savefig(outfile)


# version of above which use the RadarDisplay to return similar images.


def test_plot_netcdf_radar_rhi(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
    radar.metadata['instrument_name'] = ''
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('reflectivity_horizontal', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_netcdf_radar_ppi(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity_horizontal', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_netcdf_radar_ppi_masked(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity_horizontal', 0, ax=ax,
                     mask_tuple=('reflectivity_horizontal', 16.0))
    if outfile:
        fig.savefig(outfile)


if __name__ == "__main__":
    test_plot_netcdf_rhi('figure_plot_cfradial_rhi.png')
    test_plot_netcdf_ppi('figure_plot_cfradial_ppi.png')
    test_plot_netcdf_ppi_masked('figure_plot_cfradial_ppi_masked.png')
    test_plot_netcdf_radar_rhi('figure_plot_cfradial_rhi_radar.png')
    test_plot_netcdf_radar_ppi('figure_plot_cfradial_ppi_radar.png')
    test_plot_netcdf_radar_ppi_masked(
        'figure_plot_cfradial_ppi_masked_radar.png')



""" Unit Tests for Py-ART's graph/plot_mdv.py module. """
# execute this script to create figure_plot_mdv_*.png files.
# To compare results:
# display figure_plot_mdv_ppi.png & display figure_plot_mdv_ppi_radar.png &
# display figure_plot_mdv_rhi.png & display figure_plot_mdv_rhi_radar.png &
# display figure_plot_mdv_rhi_masked.png &
# display figure_plot_mdv_rhi_masked_radar.png &

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import pyart


def test_plot_mdv_rhi(outfile=None):
    mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_RHI_FILE)
    display = pyart.graph.MdvDisplay(mdvfile)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('DBZ_F', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_mdv_rhi_masked(outfile=None):
    mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_RHI_FILE)
    display = pyart.graph.MdvDisplay(mdvfile)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('DBZ_F', 0, mask_tuple=('DBZ_F', 16.), ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_mdv_ppi(outfile=None):
    mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_PPI_FILE)
    display = pyart.graph.MdvDisplay(mdvfile)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('DBZ_F', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


# version of above which use the RadarDisplay to return similar images.


def test_plot_mdv_radar_rhi(outfile=None):
    radar = pyart.io.read_mdv(pyart.testing.MDV_RHI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('reflectivity', 0, vmin=-16, vmax=64, ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_mdv_radar_rhi_masked(outfile=None):
    radar = pyart.io.read_mdv(pyart.testing.MDV_RHI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('reflectivity', 0, vmin=-16, vmax=64,
                     mask_tuple=('reflectivity', 16.), ax=ax)
    if outfile:
        fig.savefig(outfile)


def test_plot_mdv_radar_ppi(outfile=None):
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity', 0, vmin=-16, vmax=64, ax=ax)
    if outfile:
        fig.savefig(outfile)


if __name__ == "__main__":
    test_plot_mdv_rhi('figure_plot_mdv_rhi.png')
    test_plot_mdv_rhi_masked('figure_plot_mdv_rhi_masked.png')
    test_plot_mdv_ppi('figure_plot_mdv_ppi.png')
    test_plot_mdv_radar_rhi('figure_plot_mdv_rhi_radar.png')
    test_plot_mdv_radar_rhi_masked('figure_plot_mdv_rhi_masked_radar.png')
    test_plot_mdv_radar_ppi('figure_plot_mdv_ppi_radar.png')



"""
pyart.io.grid
=============

Reading and writing Grid objects.

.. autosummary::
    :toctree: generated/

    read_grid
    write_grid

    _read_grid_cf
    _read_grid_wrf

"""

from warnings import warn

import numpy as np
import netCDF4
import datetime

from ..core.grid import Grid
from .cfradial import _ncvar_to_dict, _create_ncvar


def read_grid(filename, exclude_fields=None):
    """
    Read a netCDF grid file

    Parameters
    ----------
    filename : str
        Filename of NetCDF grid file to read.

    Other Parameters
    ----------------
    exclude_fields : list
        A list of fields to exclude from the grid object.

    Returns
    -------
    grid : Grid
        Grid object containing gridded data.

    """

    if exclude_fields is None:
        exclude_fields = []

    ncobj = netCDF4.Dataset(filename, mode='r')

    # metadata
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])

    # axes
    axes_keys = ['time', 'time_start', 'time_end', 'base_time',
                 'time_offset', 'z_disp', 'y_disp', 'x_disp',
                 'alt', 'lat', 'lon', 'z', 'lev', 'y', 'x']
    axes = dict((k, _ncvar_to_dict(ncobj.variables[k])) for k in axes_keys
                if k in ncobj.variables)

    # read in the fields
    # determine the correct shape of the fields
    # ARM standard requires the left-most dimension to be time, so the shape
    # of the fields in the file is (1, nz, ny, nx) but the field data should
    # be shaped (nz, ny, nx) in the Grid object
    dim_keys = ['nz', 'ny', 'nx', 'z', 'y', 'x']
    field_shape = tuple([len(ncobj.dimensions[k]) for k in dim_keys
                         if k in ncobj.dimensions])
    field_shape_with_time = (1, ) + field_shape

    # check all non-axes variables, those with the correct shape
    # are added to the field dictionary, if a wrong sized field is
    # detected a warning is raised
    field_keys = [k for k in ncobj.variables if k not in axes_keys
                  and k not in exclude_fields]
    fields = {}
    for field in field_keys:
        field_dic = _ncvar_to_dict(ncobj.variables[field])
        if field_dic['data'].shape == field_shape_with_time:
            field_dic['data'].shape = field_shape
            fields[field] = field_dic
        else:
            bad_shape = field_dic['data'].shape
            warn('Field %s skipped due to incorrect shape' % (field))

    return Grid(fields, axes, metadata)


def write_grid(filename, grid, format='NETCDF4', arm_time_variables=False):
    """
    Write a Grid object to a CF-1.5 and ARM standard netcdf file

    Parameters
    ----------
    filename : str
        Filename to save grid to.
    grid : Grid
        Grid object to write.
    format : str, optional
        NetCDF format, one of 'NETCDF4', 'NETCDF4_CLASSIC',
        'NETCDF3_CLASSIC' or 'NETCDF3_64BIT'. See netCDF4 documentation for
        details.
    arm_time_variables : bool
        True to write the ARM standard time variables base_time and
        time_offset. False will not write these variables.

    """
    ncobj = netCDF4.Dataset(filename, mode='w', format=format)

    # create the time dimension
    ncobj.createDimension('time', None)

    # create additional dimensions
    grid_shape = grid.fields[grid.fields.keys()[0]]['data'].shape
    nz, ny, nx = grid_shape
    ncobj.createDimension('nz', nz)
    ncobj.createDimension('ny', ny)
    ncobj.createDimension('nx', nx)

    # axes variables
    _create_ncvar(grid.axes['time'], ncobj, 'time', ('time', ))
    _create_ncvar(grid.axes['time_end'], ncobj, 'time_end', ('time', ))
    _create_ncvar(grid.axes['time_start'], ncobj, 'time_start', ('time', ))
    _create_ncvar(grid.axes['x_disp'], ncobj, 'x_disp', ('nx', ))
    _create_ncvar(grid.axes['y_disp'], ncobj, 'y_disp', ('ny', ))
    _create_ncvar(grid.axes['z_disp'], ncobj, 'z_disp', ('nz', ))
    _create_ncvar(grid.axes['lat'], ncobj, 'lat', ('time', ))
    _create_ncvar(grid.axes['lon'], ncobj, 'lon', ('time', ))
    _create_ncvar(grid.axes['alt'], ncobj, 'alt', ('time', ))

    # create ARM time variables base_time and time_offset, if requested
    if arm_time_variables:
        time = grid.axes['time']
        dt = netCDF4.num2date(time['data'][0], time['units'])
        td = dt - datetime.datetime.utcfromtimestamp(0)
        td = td.seconds + td.days * 24 * 3600

        base_time = {
            'data': np.array([td], dtype=np.int32),
            'string': dt.strftime('%d-%b-%Y,%H:%M:%S GMT'),
            'units': 'seconds since 1970-1-1 0:00:00 0:00',
            'ancillary_variables': 'time_offset',
            'long_name': 'Base time in Epoch',
        }
        _create_ncvar(base_time, ncobj, 'base_time', ())

        time_offset = {
            'data': np.array(time['data'], dtype=np.float64),
            'long_name': 'Time offset from base_time',
            'units': time['units'].replace('T', ' ').replace('Z', ''),
            'ancillary_variables': 'time_offset',
            'calendar': 'gregorian',
        }
        _create_ncvar(time_offset, ncobj, 'time_offset', ('time', ))

    # field variables
    for field, field_dic in grid.fields.iteritems():
        # append 1, to the shape of all data to indicate the time var.
        field_dic['data'].shape = (1, ) + field_dic['data'].shape
        _create_ncvar(field_dic, ncobj, field, ('time', 'nz', 'ny', 'nx'))
        field_dic['data'].shape = field_dic['data'].shape[1:]

    # metadata
    for k, v in grid.metadata.iteritems():
        setattr(ncobj, k, v)

    ncobj.close()

    return


def _read_grid_cf(filename):
    """
    Read a CF compliant netCDF file containing a grid.

    Parameters
    ----------
    filename : str
        Filename of the netCDF file.

    Returns
    -------
    grid : Grid
        Grid object containing gridded data.

    Notes
    -----
    This function does only the most basic variable checking.  The resulting
    Grid object is most likely not writable.

    """
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables
    fields = {}
    axes = {}
    for var in ncvars:
        if len(ncvars[var].shape) > 1:
            # dimensionality of 2+ are fields variables
            fields[var] = _ncvar_to_dict(ncvars[var])
        else:
            # dimensionality of 1 are axes variables
            axes[var] = _ncvar_to_dict(ncvars[var])
    return Grid(fields, axes, {})


def _read_grid_wrf(filename):
    """
    Read a WRF netCDF file containing a grid.

    Parameters
    ----------
    filename : str
        Filename of the WRF netCDF file.

    Returns
    -------
    grid : Grid
        Grid object containing data.

    Notes
    -----
    This function does only the most basic variable checking.  The resulting
    Grid object is most likely not writable.

    """
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables
    fields = {}
    axes = {}
    for var in ncvars:
        if len(ncvars[var].shape) > 1:
            # dimensionality of 2+ are fields variables
            fields[var] = _ncvar_to_dict(ncvars[var])
        else:
            # dimensionality of 1 are axes variables
            axes[var] = _ncvar_to_dict(ncvars[var])
    return Grid(fields, axes, {})



"""
================================================
Auxiliary input and output (:mod:`pyart.aux_io`)
================================================

.. currentmodule:: pyart.aux_io

Additional classes and functions for reading and writing data from a number
of file formats.  These auxiliary input/output routines are not as well
polished as those in :mod:`pyart.io`.  They may require addition dependencies
beyond those required for a standard Py-ART install, use non-standard
function parameter and naming, are not supported by the
:py:func:`pyart.io.read` function and are not fully tested if tested at all.
Please use these at your own risk.  Bugs in these function should be
reported but fixing them will not be a priority.

.. autosummary::
    :toctree: generated/

    read_gamic
    read_pattern
    read_radx

"""

from pattern import read_pattern
from radx import read_radx
try:
    from .gamic_hdf5 import read_gamic
    from .odim_h5 import read_odim_h5
    _HDF5_AVAILABLE = True
except:
    _HDF5_AVAILABLE = False

__all__ = [s for s in dir() if not s.startswith('_')]



""" Unit Tests for Py-ART's config.py module. """

import pyart
from StringIO import StringIO


def test_debug_info():
    # test to see that something is written when _debug_info is called
    # we don't care what is written, just that something is.
    buf = StringIO()
    pyart._debug_info(buf)
    assert buf.len > 0



"""
pyart.testing.sample_files
==========================

Sample radar files in a number of formats.  Many of these files
are incomplete, they should only be used for testing, not production.

.. autosummary::
    :toctree: generated/

    MDV_PPI_FILE
    MDV_RHI_FILE
    CFRADIAL_PPI_FILE
    CFRADIAL_RHI_FILE
    CHL_RHI_FILE
    SIGMET_PPI_FILE
    SIGMET_RHI_FILE
    NEXRAD_ARCHIVE_FILE
    NEXRAD_ARCHIVE_COMPRESSED_FILE
    NEXRAD_CDM_FILE
    INTERP_SOUNDE_FILE

"""

import os

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')

MDV_PPI_FILE = os.path.join(DATA_PATH, 'example_mdv_ppi.mdv')
MDV_RHI_FILE = os.path.join(DATA_PATH, 'example_mdv_rhi.mdv')
CFRADIAL_PPI_FILE = os.path.join(DATA_PATH, 'example_cfradial_ppi.nc')
CFRADIAL_RHI_FILE = os.path.join(DATA_PATH, 'example_cfradial_rhi.nc')
CHL_RHI_FILE = os.path.join(DATA_PATH, 'example_chl_rhi.chl')
SIGMET_PPI_FILE = os.path.join(DATA_PATH, 'example_sigmet_ppi.sigmet')
SIGMET_RHI_FILE = os.path.join(DATA_PATH, 'example_sigmet_rhi.sigmet')
NEXRAD_ARCHIVE_FILE = os.path.join(DATA_PATH, 'example_nexrad_archive.bz2')
NEXRAD_ARCHIVE_COMPRESSED_FILE = os.path.join(
    DATA_PATH, 'example_nexrad_archive_compressed.ar2v')
NEXRAD_CDM_FILE = os.path.join(DATA_PATH, 'example_nexrad_cdm.bz2')
INTERP_SOUNDE_FILE = os.path.join(DATA_PATH, 'example_interpolatedsonde.cdf')
_EXAMPLE_RAYS_FILE = os.path.join(DATA_PATH, 'example_rays.npz')



"""
pyart.aux_io.odim_h5
====================

Routines for reading ODIM_H5 files.

.. autosummary::
    :toctree: generated/

    read_odim_h5

"""

import datetime

import numpy as np
import h5py

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, radar_coords_to_cart
from ..core.radar import Radar


ODIM_H5_FIELD_NAMES = {
    'TH': 'total_power',        # uncorrected reflectivity, horizontal
    'TV': 'total_power',        # uncorrected reflectivity, vertical
    'DBZH': 'reflectivity',     # corrected reflectivity, horizontal
    'DBZV': 'reflectivity',     # corrected reflectivity, vertical
    'ZDR': 'differential_reflectivity',     # differential reflectivity
    'RHOHV': 'cross_correlation_ratio',
    'LDR': 'linear_polarization_ratio',
    'PHIDP': 'differential_phase',
    'KDP': 'specific_differential_phase',
    'SQI': 'normalized_coherent_power',
    'SNR': 'signal_to_noise_ratio',
    'VRAD': 'velocity',
    'WRAD': 'spectrum_width',
    'QIND': 'quality_index',
}


def read_odim_h5(filename, field_names=None, additional_metadata=None,
                 file_field_names=False, exclude_fields=None):
    """
    Read a ODIM_H5 file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.


    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata('odim_h5', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    # open the file
    hfile = h5py.File(filename, 'r')
    odim_object = hfile['what'].attrs['object']
    if odim_object != 'PVOL':
        raise NotImplementedError(
            'object: %s not implemented.' % (odim_object))

    # determine the number of sweeps by the number of groups which
    # begin with dataset
    datasets = [k for k in hfile if k.startswith('dataset')]
    datasets.sort()
    nsweeps = len(datasets)

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    h_where = hfile['where'].attrs
    latitude['data'] = np.array([h_where['lat']], dtype='float64')
    longitude['data'] = np.array([h_where['lon']], dtype='float64')
    altitude['data'] = np.array([h_where['height']], dtype='float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata['source'] = hfile['what'].attrs['source']
    metadata['original_container'] = 'odim_h5'
    metadata['odim_conventions'] = hfile.attrs['Conventions']

    h_what = hfile['what'].attrs
    metadata['version'] = h_what['version']
    metadata['source'] = h_what['source']

    ds1_how = hfile[datasets[0]]['how'].attrs
    if 'system' in ds1_how:
        metadata['system'] = ds1_how['system']
    if 'software' in ds1_how:
        metadata['software'] = ds1_how['software']
    if 'sw_version' in ds1_how:
        metadata['sw_version'] = ds1_how['sw_version']

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    rays_per_sweep = [hfile[d]['where'].attrs['nrays'] for d in datasets]
    total_rays = sum(rays_per_sweep)
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
    seri = np.cumsum(rays_per_sweep).astype('int32') - 1
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = seri

    # sweep_number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')

    # sweep_mode
    sweep_mode = filemetadata('sweep_mode')
    sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

    # scan_type
    scan_type = 'ppi'

    # fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    sweep_el = [hfile[d]['where'].attrs['elangle'] for d in datasets]
    fixed_angle['data'] = np.array(sweep_el, dtype='float32')

    # elevation
    elevation = filemetadata('elevation')
    if 'elangles' in ds1_how:
        edata = np.empty(total_rays, dtype='float32')
        for d, start, stop in zip(datasets, ssri, seri):
            edata[start:stop+1] = hfile[d]['how'].attrs['elangles'][:]
        elevation['data'] = edata
    else:
        elevation['data'] = np.repeat(sweep_el, rays_per_sweep)

    # range
    _range = filemetadata('range')
    # check that the gate spacing is constant between sweeps
    rstart = [hfile[d]['where'].attrs['rstart'] for d in datasets]
    if any(rstart != rstart[0]):
        raise ValueError('range start changes between sweeps')
    rscale = [hfile[d]['where'].attrs['rscale'] for d in datasets]
    if any(rscale != rscale[0]):
        raise ValueError('range scale changes between sweeps')

    nbins = hfile['dataset1']['where'].attrs['nbins']
    _range['data'] = (np.arange(nbins, dtype='float32') * rscale[0] +
                      rstart[0] * 1000.)
    _range['meters_to_center_of_first_gate'] = rstart[0] * 1000.
    _range['meters_between_gates'] = float(rscale[0])

    # azimuth
    azimuth = filemetadata('azimuth')
    if ('startazA' in ds1_how) and ('stopazA' in ds1_how):
        # average between start and stop azimuth angles
        az_data = np.ones((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            startaz = hfile[dset]['how'].attrs['startazA']
            stopaz = hfile[dset]['how'].attrs['stopazA']
            sweep_az = np.angle(
                (np.exp(1.j*np.deg2rad(startaz)) +
                 np.exp(1.j*np.deg2rad(stopaz))) / 2., deg=True)
            az_data[start:stop+1] = sweep_az
        azimuth['data'] = az_data
    else:
        # assume 1 degree per ray, starting at where/a1gate
        az_data = np.ones((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            start_az = hfile[dset]['where'].attrs['a1gate']
            nrays = stop - start + 1
            sweep_az = np.fmod(start_az + np.arange(nrays), 360.)
            az_data[start:stop+1] = sweep_az
        azimuth['data'] = az_data

    # time
    _time = filemetadata('time')
    if ('startazT' in ds1_how) and ('stopazT' in ds1_how):
        # average between startazT and stopazT
        t_data = np.empty((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            t_start = hfile[dset]['how'].attrs['startazT']
            t_stop = hfile[dset]['how'].attrs['stopazT']
            t_data[start:stop+1] = (t_start + t_stop) / 2
        start_epoch = t_data.min()
        start_time = datetime.datetime.utcfromtimestamp(start_epoch)
        _time['units'] = make_time_unit_str(start_time)
        _time['data'] = t_data - start_epoch
    else:
        t_data = np.empty((total_rays, ), dtype='int32')
        # interpolate between each sweep starting and ending time
        for dset, start, stop in zip(datasets, ssri, seri):
            dset_what = hfile[dset]['what'].attrs
            start_str = dset_what['startdate'] + dset_what['starttime']
            end_str = dset_what['enddate'] + dset_what['endtime']
            start_dt = datetime.datetime.strptime(start_str, '%Y%m%d%H%M%S')
            end_dt = datetime.datetime.strptime(end_str, '%Y%m%d%H%M%S')

            time_delta = end_dt - start_dt
            delta_seconds = time_delta.seconds + time_delta.days * 3600 * 24
            rays = stop - start + 1
            sweep_start_epoch = (
                start_dt - datetime.datetime(1970, 1, 1)).total_seconds()
            t_data[start:stop+1] = (sweep_start_epoch +
                                    np.linspace(0, delta_seconds, rays))
        start_epoch = t_data.min()
        start_time = datetime.datetime.utcfromtimestamp(start_epoch)
        _time['units'] = make_time_unit_str(start_time)
        _time['data'] = (t_data - start_epoch).astype('float32')

    # fields
    fields = {}
    h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]
    odim_fields = [hfile['dataset1'][d]['what'].attrs['quantity'] for d in
                   h_field_keys]
    for odim_field, h_field_key in zip(odim_fields, h_field_keys):
        field_name = filemetadata.get_field_name(odim_field)
        if field_name is None:
            continue
        fdata = np.ma.zeros((total_rays, nbins), dtype='float32')
        start = 0
        # loop over the sweeps, copy data into correct location in data array
        for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
            sweep_data = _get_odim_h5_sweep_data(hfile[dset][h_field_key])
            fdata[start:start + rays_in_sweep] = sweep_data[:]
            start += rays_in_sweep
        # create field dictionary
        field_dic = filemetadata(field_name)
        field_dic['data'] = fdata
        field_dic['_FillValue'] = get_fillvalue()
        fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _get_odim_h5_sweep_data(group):
    """ Get ODIM_H5 sweet data from an HDF5 group. """

    # mask raw data
    what = group['what']
    raw_data = group['data'][:]

    if 'nodata' in what.attrs:
        nodata = what.attrs.get('nodata')
        data = np.ma.masked_equal(raw_data, nodata)
    else:
        data = np.ma.masked_array(raw_data)
    if 'undetect' in what.attrs:
        undetect = what.attrs.get('undetect')
        data[data == undetect] = np.ma.masked

    offset = 0.0
    gain = 1.0
    if 'offset' in what.attrs:
        offset = what.attrs.get('offset')
    if 'gain' in what.attrs:
        gain = what.attrs.get('gain')
    return data * gain + offset



"""
pyart.graph.gridmapdisplay
==========================

A class for plotting grid objects with a basemap.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    GridMapDisplay

"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
from mpl_toolkits.basemap import pyproj


class GridMapDisplay():
    """
    A class for creating plots from a grid object on top of a Basemap.

    Parameters
    ----------
    grid : Grid
        Grid with data which will be used to create plots.
    debug : bool
        True to print debugging messages, False to supress them.

    Attributes
    ----------
    grid : Grid
        Grid object.
    debug : bool
        True to print debugging messages, False to supressed them.
    proj : Proj
        Object for performing cartographic transformations specific to the
        grid.
    grid_lons : array
        Grid longitudes in degrees.
    grid_lats : array
        Grid latitudes in degress.
    basemap : Basemap
        Last plotted basemap, None when no basemap has been plotted.
    mappables : list
        List of ContourSet, etc. which have been plotted, useful
        when adding colorbars.
    fields : list
        List of fields which have been plotted.

    """

    def __init__(self, grid, debug=False):
        """ initalize the object. """
        self.grid = grid
        self.debug = debug

        # set up the projection
        lat0 = grid.axes['lat']['data'][0]
        lon0 = grid.axes['lon']['data'][0]
        self.proj = pyproj.Proj(proj='lcc', datum='NAD83',
                                lat_0=lat0, lon_0=lon0,
                                x_0=0.0, y_0=0.0)

       # determine grid latitudes and longitudes.
        x_1d = grid.axes['x_disp']['data']
        y_1d = grid.axes['y_disp']['data']
        x_2d, y_2d = np.meshgrid(x_1d, y_1d)
        self.grid_lons, self.grid_lats = self.proj(x_2d, y_2d, inverse=True)

        # set attributes
        self.mappables = []
        self.fields = []
        self.basemap = None

    def plot_basemap(self, lat_lines=None, lon_lines=None,
                     resolution='l', area_thresh=10000,
                     auto_range=True,
                     min_lon=-92, max_lon=-86, min_lat=40, max_lat=44,
                     ax=None):
        """
        Plot a basemap.

        Parameters
        ----------
        lat_lines, lon_lines : array or None
            Locations at which to draw latitude and longitude lines.
            None will use default values which are resonable for maps of
            North America.
        auto_range : bool
            True to determine map ranges from the grid_lats and grid_lons
            attribute.  False will use the min_lon, max_lon, min_lat, and
            max_lat parameters for the map range.
        min_lat, max_lat, min_lon, max_lon : float
            Latitude and longitude ranges for the map projection region in
            degrees.  These parameter are not used if auto_range is True.
        resolution : 'c', 'l', 'i', 'h', or 'f'.
            Resolution of boundary database to use. See Basemap documentation
            for details.
        area_thresh : int
            Basemap area_thresh parameter. See Basemap documentation.
        ax : axes or None.
            Axis to add the basemap to, if None the current axis is used.

        """
        # parse the parameters
        ax = self._parse_ax(ax)
        if lat_lines is None:
            lat_lines = np.arange(30, 46, 1)
        if lon_lines is None:
            lon_lines = np.arange(-110, -75, 1)

        # determine map region
        if auto_range:
            max_lat = self.grid_lats.max()
            max_lon = self.grid_lons.max()
            min_lat = self.grid_lats.min()
            min_lon = self.grid_lons.min()

        if self.debug:
            print "Maximum latitude: ", max_lat
            print "Maximum longitude: ", max_lon
            print "Minimum latitude: ", min_lat
            print "Minimum longitute: ", min_lon

        # plot the basemap
        basemap = Basemap(llcrnrlon=min_lon, llcrnrlat=min_lat,
                          urcrnrlon=max_lon, urcrnrlat=max_lat,
                          projection='mill', area_thresh=area_thresh,
                          resolution=resolution, ax=ax)
        basemap.drawcoastlines(linewidth=1.25)
        basemap.drawstates()
        basemap.drawparallels(lat_lines, labels=[True, False, False, False])
        basemap.drawmeridians(lon_lines, labels=[False, False, False, True])
        self.basemap = basemap

    def plot_grid(self, field, level=0, vmin=None, vmax=None, cmap='jet'):
        """
        Plot the grid onto the current basemap.

        Parameters
        ----------
        field : str
            Field to be plotted.
        level : int
            Index of the z level to be plotted.
        vmin, vmax : float
            Lower and upper range for the colormesh.  If either parameter is
            None, a value will be determined from the field attributes (if
            available) or the default values of -8, 64 will be used.
        cmap : str
            Matplotlib colormap name or colormap object.

        """
        # parse parameters
        vmin, vmax = self._parse_vmin_vmax(field, vmin, vmax)
        self._check_basemap()

        # plot the grid
        xd, yd = self.basemap(self.grid_lons, self.grid_lats)
        self.mappables.append(self.basemap.pcolormesh(
            xd, yd, self.grid.fields[field]['data'][level],
            vmin=vmin, vmax=vmax, cmap=cmap))
        self.fields.append(field)
        return

    def plot_crosshairs(self, lon=None, lat=None,
                        line_style='r--', linewidth=2, ax=None):
        """
        Plot crosshairs at a given longitude and latitude.

        Parameters
        ----------
        lon, lat : float
            Longitude and latitude (in degrees) where the crosshairs should
            be placed.  If None the center of the grid is used.
        line_style : str
            Matplotlib string describing the line style.
        linewidth : float
            Width of markers in points.
        ax : axes or None.
            Axis to add the crosshairs to, if None the current axis is used.

        """
        # parse the parameters
        ax = self._parse_ax(ax)
        lon, lat = self._parse_lon_lat(lon, lat)
        self._check_basemap()

        # add crosshairs.
        x_lon, y_lon = self.basemap(
            np.array([lon, lon]),
            np.array([self.basemap.latmin, self.basemap.latmax]))
        x_lat, y_lat = self.basemap(
            np.array([self.basemap.lonmin, self.basemap.lonmax]),
            np.array([lat, lat]))
        ax.plot(x_lon, y_lon, line_style, linewidth=linewidth)
        ax.plot(x_lat, y_lat, line_style, linewidth=linewidth)
        return

    def plot_latitude_slice(self, field, lon=None, lat=None,
                            vmin=None, vmax=None, cmap='jet', ax=None):
        """
        Plot a slice along a given latitude.

        Parameters
        ----------
        field : str
            Field to be plotted.
        lon, lat : float
            Longitude and latitude (in degrees) specifying the slice.  If
            None the center of the grid is used.
        vmin, vmax : float
            Lower and upper range for the colormesh.  If either parameter is
            None, a value will be determined from the field attributes (if
            available) or the default values of -8, 64 will be used.
        cmap : str
            Matplotlib colormap name or colormap object.
        ax : axes or None.
            Where to create the plot, if None the current axis is used.


        """
        # parse the parameters
        ax = self._parse_ax(ax)
        lon, lat = self._parse_lon_lat(lon, lat)
        vmin, vmax = self._parse_vmin_vmax(field, vmin, vmax)

        # plot the slice.
        x_index, y_index = self._find_nearest_grid_indices(lon, lat)
        self.mappables.append(ax.pcolormesh(
            self.grid.axes['x_disp']['data'] / 1000.0,
            self.grid.axes['z_disp']['data'] / 1000.0,
            self.grid.fields[field]['data'][:, y_index, :],
            vmin=vmin, vmax=vmax, cmap=cmap))
        self.fields.append(field)
        ax.set_ylabel('Height (km)')
        ax.set_title('Slice at ' + str(lat) + ' Latitude')
        return

    def plot_longitude_slice(self, field, lon=None, lat=None,
                             vmin=None, vmax=None, cmap='jet', ax=None):
        """
        Plot a slice along a given longitude.

        Parameters
        ----------
        field : str
            Field to be plotted.
        lon, lat : float
            Longitude and latitude (in degrees) specifying the slice.  If
            None the center of the grid is used.
        vmin, vmax : float
            Lower and upper range for the colormesh.  If either parameter is
            None, a value will be determined from the field attributes (if
            available) or the default values of -8, 64 will be used.
        cmap : str
            Matplotlib colormap name or colormap object.
        ax : axes or None.
            Where to create the plot, if None the current axis is used.

        """
        # parse the parameters
        ax = self._parse_ax(ax)
        lon, lat = self._parse_lon_lat(lon, lat)
        vmin, vmax = self._parse_vmin_vmax(field, vmin, vmax)

        # plot the slice
        x_index, y_index = self._find_nearest_grid_indices(lon, lat)
        self.mappables.append(ax.pcolormesh(
            self.grid.axes['y_disp']['data'] / 1000.0,
            self.grid.axes['z_disp']['data'] / 1000.0,
            self.grid.fields[field]['data'][:, :, x_index],
            vmin=vmin, vmax=vmax, cmap=cmap))
        self.fields.append(field)
        ax.set_ylabel('Height (km)')
        ax.set_title('Slice at ' + str(lon) + ' Longitude')
        return

    def plot_colorbar(self, mappable=None, orientation='horizontal',
                      label=None, cax=None):
        """
        Plot a colorbar to an axis.

        Parameters
        ----------
        mappable : ContourSet, etc.
            Matplotlib object to create colorbar for.  None will use
            the last mappable plotted.
        orientation : 'vertical' or 'horizontal'
            Colorbar orientation.
        label : str
            Colorbar label.  None will use: 'long_name' ('units') for
            the last field plotted or '' if the field does not have these
            keys.
        cax : axis
            Axes object into which the colorbar will be drawn.
            None is also allowed.

        """
        if mappable is None:
            if len(self.mappables) == 0:
                raise ValueError('mappable must be specified.')
            else:
                mappable = self.mappables[-1]

        if label is None:
            if len(self.fields) == 0:
                raise ValueError('field must be specified.')
            field = self.grid.fields[self.fields[-1]]
            if 'long_name' in field and 'units' in field:
                label = field['long_name'] + '(' + field['units'] + ')'
            else:
                label = ''

        # plot the colorbar and set the label.
        cb = plt.colorbar(cax=cax, mappable=mappable, orientation=orientation)
        cb.set_label(label)

        return

    def _check_basemap(self):
        """ Check that basemap is not None, raise ValueError if it is. """
        if self.basemap is None:
            raise ValueError('no basemap plotted')

    def _find_nearest_grid_indices(self, lon, lat):
        """
        Find the nearest x, y grid indices for a given latitude and longitude.
        """
        lon, lat = self._parse_lon_lat(lon, lat)
        x_cut, y_cut = self.proj(lon, lat)

        if self.debug:
            print "x_cut: ", x_cut,
            print "y_cut: ", y_cut

        x_index = np.abs(self.grid.axes['x_disp']['data'] - x_cut).argmin()
        y_index = np.abs(self.grid.axes['y_disp']['data'] - y_cut).argmin()

        if self.debug:
            print "x_index", x_index
            print "y_index", y_index

        return x_index, y_index

    def _parse_vmin_vmax(self, field, vmin, vmax):
        """ Parse vmin and vmax parameters. """
        if vmin is None:
            if 'valid_min' in self.grid.fields[field]:
                vmin = self.grid.fields[field]['valid_min']
            else:
                vmin = -8

        if vmax is None:
            if 'valid_max' in self.grid.fields[field]:
                vmax = self.grid.fields[field]['valid_max']
            else:
                vmax = 64
        return vmin, vmax

    def _parse_lon_lat(self, lon, lat):
        """ Parse lat and lon parameters """
        if lat is None:
            lat = self.grid.axes['lat']['data'][0]
        if lon is None:
            lon = self.grid.axes['lon']['data'][0]
        return lon, lat

    def _parse_ax(self, ax):
        """ Parse the ax parameter. """
        if ax is None:
            return plt.gca()
        return ax



"""
=====================================================
Dealias doppler velocities using the FourDD algorithm
=====================================================

In this example doppler velocities are dealiased using the Univ. of
Washington FourDD algorithm implemented in Py-ART.  Sounding data is
used for the initial condition of the dealiasing.

"""
print __doc__

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import netCDF4
import pyart

SOND_NAME = 'sgpinterpolatedsondeC1.c1.20110510.000000.cdf'
RADAR_NAME = '095636.mdv'

# read in the data
radar = pyart.io.read_mdv(RADAR_NAME)

# find and extract sonde data
target = netCDF4.num2date(radar.time['data'][0], radar.time['units'])
interp_sounde = netCDF4.Dataset(SOND_NAME)
t = pyart.correct.find_time_in_interp_sonde(interp_sounde, target)
height, speed, direction = t

# perform dealiasing
dealias_data = pyart.correct.dealias_fourdd(
    radar, sounding_heights=height * 1000.0, sounding_wind_speeds=speed,
    sounding_wind_direction=direction)
radar.add_field('corrected_velocity', dealias_data)

# create a plot of the first and sixth sweeps
fig = plt.figure(figsize=(15, 10))
ax1 = fig.add_subplot(221)
display = pyart.graph.RadarDisplay(radar)
display.plot('velocity', 0, vmin=-16, vmax=16, ax=ax1,
             colorbar_label='', title='Raw Doppler Velocity, First Sweep')

ax2 = fig.add_subplot(222)
display.plot('corrected_velocity', 0, vmin=-40, vmax=40,
             colorbar_label='', ax=ax2,
             title='Corrected Doppler Velocity, First Sweep')

ax3 = fig.add_subplot(223)
display = pyart.graph.RadarDisplay(radar)
display.plot('velocity', 5, vmin=-16, vmax=16,
             colorbar_label='', ax=ax3,
             title='Raw Doppler Velocity, Sixth Sweep')

ax4 = fig.add_subplot(224)
display.plot_ppi('corrected_velocity', 5, vmin=-40, vmax=40,
                 colorbar_label='', ax=ax4,
                 title='Corrected Doppler Velocity, Sixth Sweep')
plt.suptitle('Velocity dealiasing using Py-ART', fontsize=16)
plt.show()



"""
pyart.io.common
===============

Input/output routines common to many file formats.

.. autosummary::
    :toctree: generated/

    dms_to_d
    stringarray_to_chararray
    radar_coords_to_cart
    make_time_unit_str

"""

import numpy as np
import netCDF4


def dms_to_d(dms):
    """ Degrees, minutes, seconds to degrees """
    return dms[0] + (dms[1] + dms[2] / 60.0) / 60.0


def stringarray_to_chararray(arr, numchars=None):
    """
    Convert an string array to a character array with one extra dimension.

    Parameters
    ----------
    arr : array
        Array with numpy dtype 'SN', where N is the number of characters
        in the string.

    numchars : int
        Number of characters used to represent the string.  If numchar > N
        the results will be padded on the right with blanks.  The default,
        None will use N.

    Returns
    -------
    chararr : array
        Array with dtype 'S1' and shape = arr.shape + (numchars, ).

    """
    carr = netCDF4.stringtochar(arr)
    if numchars is None:
        return carr

    arr_numchars = carr.shape[-1]
    if numchars <= arr_numchars:
        raise ValueError('numchars must be >= %i' % (arr_numchars))
    chararr = np.zeros(arr.shape + (numchars, ), dtype='S1')
    chararr[..., :arr_numchars] = carr[:]
    return chararr


# XXX move this to another module
def radar_coords_to_cart(rng, az, ele, debug=False):
    """
    Calculate Cartesian coordinate from radar coordinates

    Parameters
    ----------
    rng : array
        Distances to the center of the radar gates (bins) in kilometers.
    az : array
        Azimuth angle of the radar in degrees.
    ele : array
        Elevation angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    The calculation for Cartesian coordinate is adapted from equations
    2.28(b) and 2.28(c) of Doviak and Zrnic [1]_ assuming a
    standard atmosphere (4/3 Earth's radius model).

    .. math::

        z = \\sqrt{r^2+R^2+r*R*sin(\\theta_e)} - R

        s = R * arcsin(\\frac{r*cos(\\theta_e)}{R+z})

        x = s * sin(\\theta_a)

        y = s * cos(\\theta_a)

    Where r is the distance from the radar to the center of the gate,
    :math:`\\theta_a` is the azimuth angle, :math:`\\theta_e` is the
    elevation angle, s is the arc length, and R is the effective radius
    of the earth, taken to be 4/3 the mean radius of earth (6371 km).

    References
    ----------
    .. [1] Doviak and Zrnic, Doppler Radar and Weather Observations, Second
        Edition, 1993, p. 21.

    """
    theta_e = ele * np.pi / 180.0       # elevation angle in radians.
    theta_a = az * np.pi / 180.0        # azimuth angle in radians.
    R = 6371.0 * 1000.0 * 4.0 / 3.0     # effective radius of earth in meters.
    r = rng * 1000.0                    # distances to gates in meters.

    z = (r ** 2 + R ** 2 + 2.0 * r * R * np.sin(theta_e)) ** 0.5 - R
    s = R * np.arcsin(r * np.cos(theta_e) / (R + z))  # arc length in m.
    x = s * np.sin(theta_a)
    y = s * np.cos(theta_a)
    return x, y, z


def make_time_unit_str(dtobj):
    """ Return a time unit string from a datetime object. """
    return "seconds since " + dtobj.strftime("%Y-%m-%dT%H:%M:%SZ")



""" Unit Tests for Py-ART's core/radar.py module. """

import sys
from io import BytesIO
import inspect

import numpy as np
from numpy.testing import assert_raises
import pyart


def test_rays_per_sweep_attribute():
    radar = pyart.testing.make_target_radar()
    rays_per_sweep = radar.rays_per_sweep
    assert isinstance(rays_per_sweep, dict)
    assert rays_per_sweep['data'].shape == (1, )
    assert rays_per_sweep['data'][0] == 360


def test_iterators():
    radar = pyart.testing.make_empty_ppi_radar(30, 20, 5)
    radar.fields['reflectivity'] = {
        'data': np.zeros((100, 30), dtype=np.float32)}

    starts = [0, 20, 40, 60, 80]
    ends = [19, 39, 59, 79, 99]
    starts_ends = [(s, e) for s, e in zip(starts, ends)]

    assert inspect.isgenerator(radar.iter_start())
    assert [s for s in radar.iter_start()] == starts

    assert inspect.isgenerator(radar.iter_end())
    assert [s for s in radar.iter_end()] == ends

    assert inspect.isgenerator(radar.iter_start_end())
    assert [s for s in radar.iter_start_end()] == starts_ends

    assert inspect.isgenerator(radar.iter_slice())
    for s, start, end in zip(radar.iter_slice(), starts, ends):
        assert s.start == start
        assert s.stop == end + 1
        assert s.step is None

    assert inspect.isgenerator(radar.iter_field('reflectivity'))
    for d in radar.iter_field('reflectivity'):
        assert d.shape == (20, 30)
        assert d.dtype == np.float32
    assert_raises(KeyError, radar.iter_field, 'foobar')

    assert inspect.isgenerator(radar.iter_azimuth())
    for d in radar.iter_azimuth():
        assert d.shape == (20, )

    assert inspect.isgenerator(radar.iter_elevation())
    for d in radar.iter_elevation():
        assert d.shape == (20, )


def test_get_methods():
    radar = pyart.testing.make_empty_ppi_radar(30, 20, 5)
    radar.fields['reflectivity'] = {
        'data': np.zeros((100, 30), dtype=np.float32)}

    assert radar.get_start(0) == 0
    assert radar.get_start(1) == 20
    assert_raises(IndexError, radar.get_start, -1)
    assert_raises(IndexError, radar.get_start, 20)

    assert radar.get_end(0) == 19
    assert radar.get_end(1) == 39
    assert_raises(IndexError, radar.get_end, -1)
    assert_raises(IndexError, radar.get_end, 20)

    assert radar.get_start_end(0) == (0, 19)
    assert radar.get_start_end(1) == (20, 39)
    assert_raises(IndexError, radar.get_start_end, -1)
    assert_raises(IndexError, radar.get_start_end, 20)

    assert radar.get_slice(0) == slice(0, 20)
    assert radar.get_slice(1) == slice(20, 40)
    assert_raises(IndexError, radar.get_slice, -1)
    assert_raises(IndexError, radar.get_slice, 20)

    data = radar.get_field(0, 'reflectivity')
    assert data.shape == (20, 30)
    assert data.dtype == np.float32
    data = radar.get_field(1, 'reflectivity')
    assert data.shape == (20, 30)
    assert data.dtype == np.float32
    assert_raises(KeyError, radar.get_field, 0, 'foobar')
    assert_raises(IndexError, radar.get_field, -1, 'reflectivity')
    assert_raises(IndexError, radar.get_field, 20, 'reflectivity')

    assert radar.get_azimuth(0).shape == (20, )
    assert_raises(IndexError, radar.get_azimuth, -1)
    assert_raises(IndexError, radar.get_azimuth, 20)

    assert radar.get_elevation(0).shape == (20, )
    assert_raises(IndexError, radar.get_elevation, -1)
    assert_raises(IndexError, radar.get_elevation, 20)


def test_extract_sweeps():
    radar = pyart.testing.make_empty_ppi_radar(100, 360, 3)
    radar.fields['reflectivity'] = {'data': np.zeros((1080, 100))}
    radar.fields['velocity'] = {'data': np.zeros((1080, 100))}

    eradar = radar.extract_sweeps([0, 2])

    # extracted radar should have 720 rays, 2 sweeps, 100 gates
    assert eradar.time['data'].shape == (720, )
    assert eradar.range['data'].shape == (100, )

    assert eradar.metadata['instrument_name'] == 'fake_radar'
    assert eradar.scan_type == 'ppi'

    assert eradar.latitude['data'].shape == (1, )
    assert eradar.longitude['data'].shape == (1, )
    assert eradar.altitude['data'].shape == (1, )
    assert eradar.altitude_agl is None

    assert eradar.sweep_number['data'].shape == (2, )
    assert eradar.sweep_mode['data'].shape == (2, )
    assert eradar.fixed_angle['data'].shape == (2, )
    assert eradar.sweep_start_ray_index['data'].shape == (2, )
    assert eradar.sweep_end_ray_index['data'].shape == (2, )
    assert eradar.target_scan_rate is None

    assert eradar.azimuth['data'].shape == (720, )
    assert eradar.elevation['data'].shape == (720, )
    assert eradar.scan_rate is None
    assert eradar.antenna_transition is None

    assert eradar.instrument_parameters is None
    assert eradar.radar_calibration is None

    assert eradar.ngates == 100
    assert eradar.nrays == 720
    assert eradar.nsweeps == 2

    assert eradar.fields['reflectivity']['data'].shape == (720, 100)
    assert eradar.fields['velocity']['data'].shape == (720, 100)


def test_extract_sweeps_extra():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt': {'data': np.zeros((108, ))},
        'prt_mode': {'data': np.array(['fixed'] * 3)},
        'radar_antenna_gain_h': {'data': np.array(0)},
    }

    radar.radar_calibration = {
        'r_calib_index': {'data': np.zeros((108, ))},
        'r_calib_time': {'data': np.zeros((8, ))}
    }

    eradar = radar.extract_sweeps([0, 2])

    instr = eradar.instrument_parameters
    assert instr['prt']['data'].shape == (72, )
    assert instr['prt_mode']['data'].shape == (2, )
    assert instr['radar_antenna_gain_h']['data'].shape == ()

    calib = eradar.radar_calibration
    assert calib['r_calib_index']['data'].shape == (72, )
    assert calib['r_calib_time']['data'].shape == (8, )


def test_extract_sweeps_errors():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 2)
    assert_raises(ValueError, radar.extract_sweeps, [0, 2])
    assert_raises(ValueError, radar.extract_sweeps, [-1, 1])


def test_radar_creation():
    radar = pyart.testing.make_target_radar()
    assert isinstance(radar, pyart.core.Radar)


def test_add_field():
    radar = pyart.testing.make_target_radar()
    dic = {'data': np.zeros((360, 50)), 'standard_name': 'test'}
    radar.add_field('test', dic)
    assert 'test' in radar.fields
    assert 'data' in radar.fields['test']
    assert radar.fields['test']['standard_name'] == 'test'


def test_add_field_errors():
    radar = pyart.testing.make_target_radar()

    assert_raises(ValueError, radar.add_field, 'reflectivity', {})

    dic = {'dat': np.zeros((360, 50)), 'standard_name': 'test'}
    assert_raises(KeyError, radar.add_field, 'test', dic)

    dic = {'data': np.zeros((360, 49)), 'standard_name': 'test'}
    assert_raises(ValueError, radar.add_field, 'test', dic)


def test_add_field_like():
    radar = pyart.testing.make_target_radar()
    data = np.zeros((360, 50))
    radar.add_field_like('reflectivity', 'test', data)
    assert 'test' in radar.fields
    assert 'data' in radar.fields['test']
    assert radar.fields['test']['units'] == 'dBZ'


def test_add_field_like_bug():
    # tests for bug where adding a field over-writes 'like' field
    # data/metadata.
    radar = pyart.testing.make_target_radar()
    data = np.ones((360, 50))
    radar.add_field_like('reflectivity', 'test', data)
    radar.fields['test']['units'] = 'fake'

    # check field added
    assert radar.fields['test']['units'] == 'fake'
    assert radar.fields['test']['data'][0, 0] == 1

    # check original field
    assert radar.fields['reflectivity']['units'] == 'dBZ'
    assert radar.fields['reflectivity']['data'][0, 0] == 0


def test_add_field_like_errors():
    radar = pyart.testing.make_target_radar()
    assert_raises(ValueError, radar.add_field_like, 'foo', 'bar', [])


def test_info_levels():
    for level in ['standard', 's', 'compact', 'c', 'full', 'f']:
        yield check_info, level


def test_info_nonstandard():
    # build a non-standard radar object for testing all paths in info
    radar = pyart.testing.make_target_radar()
    radar.fields['reflectivity']['data'] = [1, 2, 3, 4]
    radar.instrument_parameters = {'foobar': {'data': [1, 2], 'bar': 'foo'}}
    radar.radar_calibration = {'foobar': {'data': [1, 2], 'bar': 'foo'}}
    check_info('standard', radar)


def check_info(level, radar=None):
    out = BytesIO()
    get_info(level, out, radar)
    # don't check the output, just that something was printed.
    assert len(out.getvalue()) != 0


def get_info(level='standard', out=sys.stdout, radar=None):
    if radar is None:
        radar = pyart.testing.make_target_radar()
    radar.info(level, out)


def test_info_errors():
    assert_raises(ValueError, check_info, 'foo')


def test_is_vpt():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    assert not pyart.core.is_vpt(radar)
    pyart.core.to_vpt(radar)
    assert pyart.core.is_vpt(radar)


def test_to_vpt():
    # single scan
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt_mode': {'data': np.array(['fixed'] * 3)}
    }
    pyart.core.to_vpt(radar)
    assert pyart.core.is_vpt(radar)
    assert radar.nsweeps == 1
    assert radar.azimuth['data'][10] == 0.0
    assert radar.elevation['data'][0] == 90.0
    assert len(radar.instrument_parameters['prt_mode']['data']) == 1

    # multiple scans
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt_mode': {'data': np.array(['fixed'] * 3)}
    }
    pyart.core.to_vpt(radar, False)
    assert pyart.core.is_vpt(radar)
    assert radar.nsweeps == 108
    assert radar.azimuth['data'][10] == 10.0
    assert radar.elevation['data'][0] == 90.0
    assert len(radar.instrument_parameters['prt_mode']['data']) == 108



"""
========================================
Radar Corrections (:mod:`pyart.correct`)
========================================

.. currentmodule:: pyart.correct

Py-ART has the ability to perform a number of common corrections on radar
moments and data.

.. autosummary::
    :toctree: generated/

    dealias_fourdd
    dealias_unwrap_phase
    dealias_region_based
    calculate_attenuation
    phase_proc_lp
    GateFilter
    moment_based_gate_filter
    find_time_in_interp_sonde

"""

try:
    from .dealias import dealias_fourdd, find_time_in_interp_sonde
except ImportError:
    pass
from .attenuation import calculate_attenuation
from .phase_proc import phase_proc_lp
from .filters import GateFilter, moment_based_gate_filter
from .unwrap import dealias_unwrap_phase
from .region_dealias import dealias_region_based

__all__ = [s for s in dir() if not s.startswith('_')]



"""
========================================
Radar Retrievals (:mod:`pyart.retrieve`)
========================================

.. currentmodule:: pyart.retrieve

Functions for performing radar retrievals.

.. autosummary::
    :toctree: generated/

    steiner_conv_strat

"""

try:
    from .echo_class import steiner_conv_strat
    _F90_EXTENSIONS_AVAILABLE = True
except:
    _F90_EXTENSIONS_AVAILABLE = False

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.graph.radardisplay_airborne
=================================

Class for creating plots from Airborne Radar objects.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    RadarDisplay_Airborne

"""

import matplotlib.pyplot as plt
import numpy as np
import netCDF4
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .common import corner_to_point, radar_coords_to_cart
from .coord_transform import radar_coords_to_cart_track_relative
from .coord_transform import radar_coords_to_cart_earth_relative


class RadarDisplay_Airborne:
    """
    A display object for creating plots from data in a airborne radar object.

    Parameters
    ----------
    radar : Radar
        Radar object to use for creating plots, should be an airborne radar.
    shift : (float, float)
        Shifts in km to offset the calculated x and y locations.

    Attributes
    ----------
    plots : list
        List of plots created.
    plot_vars : list
        List of fields plotted, order matches plot list.
    cbs : list
        List of colorbars created.
    radar_name : str
        Name of radar.
    origin : str
        'Origin' or 'Radar'.
    shift : (float, float)
        Shift in meters.
    x, y, z : array
        Cartesian location of a sweep in meters.
    loc : (float, float)
        Latitude and Longitude of radar in degrees.
    time_begin : datetime
        Beginning time of first radar scan.
    starts : array
        Starting ray index for each sweep.
    ends : array
        Ending ray index for each sweep.
    fields : dict
        Radar fields.
    scan_type : str
        Scan type.
    ranges : array
        Gate ranges in meters.
    azimuths : array
        Azimuth angle in degrees.
    elevations : array
        Elevations in degrees.
    fixed_angle : array
        Scan angle in degrees.
    rotation : array
        Rotation angle in degrees.
    roll : array
        Roll angle in degrees.
    drift : array
        Drift angle in degrees.
    tilt : array
        Tilt angle in degrees.
    heading : array
        Heading angle in degrees.
    pitch : array
        Pitch angle in degrees.
    altitude : array
        Altitude angle in meters.


    """
    # TODO this class could likely be made a subclass of RadarDisplay.

    def __init__(self, radar, shift=(0.0, 0.0)):
        """ Initialize the object. """
        # populate attributes from radar object
        self.fields = radar.fields
        self.scan_type = radar.scan_type
        self.ranges = radar.range['data']
        self.azimuths = radar.azimuth['data']
        self.elevations = radar.elevation['data']
        self.fixed_angle = radar.fixed_angle['data'][0]
        self.rotation = radar.rotation['data']
        self.roll = radar.roll['data']
        self.drift = radar.drift['data']
        self.tilt = radar.tilt['data']
        self.heading = radar.heading['data']
        self.pitch = radar.pitch['data']
        self.altitude = radar.altitude['data']
        if 'instrument_name' in radar.metadata:
            self.radar_name = radar.metadata['instrument_name']
        else:
            self.radar_name = ''

        # origin
        if shift != (0.0, 0.0):
            self.origin = 'origin'
        else:
            self.origin = 'radar'

        # x, y, z attributes: cartesian location for a sweep in km.
        self.shift = shift

        if radar.metadata['platform_type'] == 'aircraft_belly':
            rg, azg = np.meshgrid(self.ranges, self.azimuths)
            rg, eleg = np.meshgrid(self.ranges, self.elevations)
            self.x, self.y, self.z = radar_coords_to_cart(
                rg / 1000.0, azg, eleg)
            self.x = self.x + self.shift[0]
            self.y = self.y + self.shift[1]
        else:
            rg, rotg = np.meshgrid(self.ranges, self.rotation)
            rg, rollg = np.meshgrid(self.ranges, self.roll)
            rg, driftg = np.meshgrid(self.ranges, self.drift)
            rg, tiltg = np.meshgrid(self.ranges, self.tilt)
            rg, pitchg = np.meshgrid(self.ranges, self.pitch)
            self.x, self.y, self.z = radar_coords_to_cart_track_relative(
                rg / 1000.0, rotg, rollg, driftg, tiltg, pitchg)
            self.x = self.x + self.shift[0]
            self.y = self.y + self.shift[1]

        # radar location in latitude and longitude
        middle_lat = int(radar.latitude['data'].shape[0] / 2)
        middle_lon = int(radar.longitude['data'].shape[0] / 2)
        lat = float(radar.latitude['data'][middle_lat])
        lon = float(radar.longitude['data'][middle_lon])
        self.loc = (lat, lon)

        # datetime object describing first sweep time
        times = radar.time['data'][0]
        units = radar.time['units']
        calendar = radar.time['calendar']
        self.time_begin = netCDF4.num2date(times, units, calendar)

        # sweep start and end indices
        self.starts = radar.sweep_start_ray_index['data']
        self.ends = radar.sweep_end_ray_index['data']

        # list to hold plots, plotted fields and plotted colorbars
        self.plots = []
        self.plot_vars = []
        self.cbs = []

    ####################
    # Plotting methods #
    ####################

    def plot(self, field, **kwargs):
        """
        Create a plot appropiate for the radar.

        This function calls the plotting function corresponding to
        the scan_type of the radar.  Additional keywords can be passed to
        customize the plot, see the appropiate plot function for the
        allowed keywords.

        Parameters
        ----------
        field : str
            Field to plot.

        See Also
        --------
        plot_ppi : Plot a PPI scan
        plot_sweep_grid : Plot a RHI or VPT scan

        """
        if self.scan_type == 'ppi':
            self.plot_sweep_ppi(field, **kwargs)
        elif self.scan_type == 'rhi':
            self.plot_sweep_grid(field, **kwargs)
        elif self.scan_type == 'vpt':
            self.plot_sweep_grid(field, **kwargs)
        else:
            raise ValueError('unknown scan_type % s' % (self.scan_type))
        return

    def plot_ray(self, field, ray, format_str='k-', mask_tuple=None,
                 ray_min=None, ray_max=None, mask_outside=False, title=None,
                 title_flag=True, axislabels=(None, None),
                 axislabels_flag=True, ax=None, fig=None):
        """
        Plot a single ray.

        Parameters
        ----------
        field : str
            Field to plot.
        ray : int
            Ray number to plot.

        Other Parameters
        ----------------
        format_str : str
            Format string defining the line style and marker.
        mask_tuple : (str, float)
            Tuple containing the field name and value below which to mask
            field prior to plotting, for example to mask all data where
            NCP < 0.5 set mask_tuple to ['NCP', 0.5]. None performs no masking.
        ray_min : float
            Minimum ray value, None for default value, ignored if mask_outside
            is False.
        ray_max : float
            Maximum ray value, None for default value, ignored if mask_outside
            is False.
        mask_outside : bool
            True to mask data outside of vmin, vmax.  False performs no
            masking.
        title : str
            Title to label plot with, None to use default title generated from
            the field and ray parameters. Parameter is ignored if title_flag
            is False.
        title_flag : bool
            True to add a title to the plot, False does not add a title.
        axislabels : (str, str)
            2-tuple of x-axis, y-axis labels.  None for either label will use
            the default axis label.  Parameter is ignored if axislabels_flag is
            False.
        axislabel_flag : bool
            True to add label the axes, False does not label the axes.
        ax : Axis
            Axis to plot on. None will use the current axis.
        fig : Figure
            Figure to add the colorbar to. None will use the current figure.

        """
        # parse parameters
        ax, fig = self._parse_ax_fig(ax, fig)

        # get the data and mask
        data = self._get_ray_data(field, ray, mask_tuple)

         # mask the data where outside the limits
        if mask_outside:
            data = np.ma.masked_outside(data, ray_min, ray_max)

        # plot the data
        line, = ax.plot(self.ranges / 1000., data, format_str)

        if title_flag:
            self._set_ray_title(field, ray, title, ax)

        if axislabels_flag:
            self._label_axes_ray(axislabels, field, ax)

        # add plot and field to attribute lists
        self.plots.append(line)
        self.plot_vars.append(field)

    def plot_sweep_grid(self, field, mask_tuple=None, vmin=None, vmax=None,
                        cmap='jet', mask_outside=True, title=None,
                        title_flag=True, axislabels=(None, None),
                        axislabels_flag=True, colorbar_flag=True,
                        colorbar_label=None, colorbar_orient=None,
                        ax=None, fig=None):
        """
        Plot a sweep as a grid.

        Parameters
        ----------
        field : str
            Field to plot.

        Other Parameters
        ----------------
        mask_tuple : (str, float)
            Tuple containing the field name and value below which to mask
            field prior to plotting, for example to mask all data where
            NCP < 0.5 set mask_tuple to ['NCP', 0.5]. None performs no masking.
        vmin : float
            Luminance minimum value, None for default value.
        vmax : float
            Luminance maximum value, None for default value.
        cmap : str
            Matplotlib colormap name.
        mask_outside : bool
            True to mask data outside of vmin, vmax.  False performs no
            masking.
        title : str
            Title to label plot with, None to use default title generated from
            the field and sweep parameters. Parameter is ignored if title_flag
            is False.
        title_flag : bool
            True to add a title to the plot, False does not add a title.
        axislabels : (str, str)
            2-tuple of x-axis, y-axis labels.  None for either label will use
            the default axis label.  Parameter is ignored if axislabels_flag is
            False.
        axislabel_flag : bool
            True to add label the axes, False does not label the axes.
        colorbar_flag : bool
            True to add a colorbar with label to the axis.  False leaves off
            the colorbar.
        colorbar_label : str
            Colorbar label, None will use a default label generated from the
            field information.
        colorbar_orient : str
            Colorbar orientation, None will use default orientation of
            vertical.
        ax : Axis
            Axis to plot on. None will use the current axis.
        fig : Figure
            Figure to add the colorbar to. None will use the current figure.

        """
        # parse parameters
        ax, fig = self._parse_ax_fig(ax, fig)
        vmin, vmax = self._parse_vmin_vmax(field, vmin, vmax)

        # get data for the plot
        data = self._get_data(field, mask_tuple=mask_tuple)
        x, z = self._get_x_z(field)

        # mask the data where outside the limits
        if mask_outside:
            data = np.ma.masked_outside(data, vmin, vmax)

        # plot the data
        pm = ax.pcolormesh(x, z, data, vmin=vmin, vmax=vmax, cmap=cmap)

        # Set the aspcet ratio
        # ax.axis('scaled')

        if title_flag:
            self._set_title(field, title, ax)

        if axislabels_flag:
            self._label_axes_sweep(axislabels, ax)

        # add plot and field to lists
        self.plots.append(pm)
        self.plot_vars.append(field)

        # colorbar options
        if colorbar_flag:
            self.plot_colorbar(mappable=pm, label=colorbar_label,
                               orient=colorbar_orient,
                               field=field, ax=ax, fig=fig)

    def plot_ppi(self, field, sweep=0, mask_tuple=None, vmin=None, vmax=None,
                 cmap='jet', mask_outside=True, title=None, title_flag=True,
                 axislabels=(None, None), axislabels_flag=True,
                 colorbar_flag=True, colorbar_label=None, ax=None, fig=None):
        """
        Plot a PPI.

        Parameters
        ----------
        field : str
            Field to plot.
        sweep : int, optional
            Sweep number to plot.

        Other Parameters
        ----------------
        mask_tuple : (str, float)
            Tuple containing the field name and value below which to mask
            field prior to plotting, for example to mask all data where
            NCP < 0.5 set mask_tuple to ['NCP', 0.5]. None performs no masking.
        vmin : float
            Luminance minimum value, None for default value.
        vmax : float
            Luminance maximum value, None for default value.
        cmap : str
            Matplotlib colormap name.
        mask_outside : bool
            True to mask data outside of vmin, vmax.  False performs no
            masking.
        title : str
            Title to label plot with, None to use default title generated from
            the field and sweep parameters. Parameter is ignored if title_flag
            is False.
        title_flag : bool
            True to add a title to the plot, False does not add a title.
        axislabels : (str, str)
            2-tuple of x-axis, y-axis labels.  None for either label will use
            the default axis label.  Parameter is ignored if axislabels_flag is
            False.
        axislabel_flag : bool
            True to add label the axes, False does not label the axes.
        colorbar_flag : bool
            True to add a colorbar with label to the axis.  False leaves off
            the colorbar.
        colorbar_label : str
            Colorbar label, None will use a default label generated from the
            field information.
        ax : Axis
            Axis to plot on. None will use the current axis.
        fig : Figure
            Figure to add the colorbar to. None will use the current figure.

        """
        # parse parameters
        ax, fig = self._parse_ax_fig(ax, fig)
        vmin, vmax = self._parse_vmin_vmax(field, vmin, vmax)

        # get data for the plot
        data = self._get_data(field, mask_tuple)
        x, y = self._get_x_y(field)

        # mask the data where outside the limits
        if mask_outside:
            data = np.ma.masked_outside(data, vmin, vmax)

        # plot the data
        pm = ax.pcolormesh(x, y, data, vmin=vmin, vmax=vmax, cmap=cmap)

        if title_flag:
            self._set_title(field, title, ax)

        if axislabels_flag:
            self._label_axes_ppi(axislabels, ax)

        # add plot and field to lists
        self.plots.append(pm)
        self.plot_vars.append(field)

        if colorbar_flag:
            self.plot_colorbar(mappable=pm, label=colorbar_label,
                               field=field, ax=ax, fig=fig)

    def plot_range_rings(self, range_rings, ax=None, col=None, ls=None,
                         lw=None):
        """
        Plot a series of range rings.

        Parameters
        ----------
        range_rings : list
            List of locations in km to draw range rings.
        ax : Axis
            Axis to plot on.  None will use the current axis.
        col : str or value
            Color to use for range rings.
        ls : str
            Linestyle to use for range rings.

        """
        ax = self._parse_ax(ax)
        for range_ring_location_km in range_rings:
            self.plot_range_ring(range_ring_location_km, ax=ax, col=col,
                                 ls=ls, lw=lw)

    def plot_range_ring(self, range_ring_location_km, npts=100, ax=None,
                        col=None, ls=None, lw=None):
        """
        Plot a single range ring.

        Parameters
        ----------
        range_ring_location_km : float
            Location of range ring in km.
        npts: int
            Number of points in the ring, higher for better resolution.
        ax : Axis
            Axis to plot on.  None will use the current axis.
        col : str or value
            Color to use for range rings.
        ls : str
            Linestyle to use for range rings.

        """
        ax = self._parse_ax(ax)
        theta = np.linspace(0, 2 * np.pi, npts)
        r = np.ones([npts], dtype=np.float32) * range_ring_location_km
        x = r * np.sin(theta)
        y = r * np.cos(theta)
        if lw is None:
            lw = 2
        if ls is None:
            ls = '-'
        if col is None:
            col = 'k'
        ax.plot(x, y, c=col, ls=ls, lw=lw)

    def plot_grid_lines(self, ax=None, col=None, ls=None, lw=None):
        """
        Plot grid lines.

        Parameters
        ----------
        ax : Axis
            Axis to plot on.  None will use the current axis.
        col : str or value
            Color to use for grid lines.
        ls : str
            Linestyle to use for grid lines.

        """
        ax = self._parse_ax(ax)
        if lw is None:
            lw = 2
        if ls is None:
            ls = ':'
        if col is None:
            col = 'k'
        ax.grid(c=col, ls=ls)

    def plot_labels(self, labels, locations, symbols='r+', text_color='k',
                    ax=None):
        """
        Plot symbols and labels at given locations.

        Parameters
        ----------
        labels : list of str
            List of labels to place just above symbols.
        locations : list of 2-tuples
            List of latitude, longitude (in degrees) tuples at which symbols
            will be place.  Labels are placed just above the symbols.
        symbols : list of str or str
            List of matplotlib color+marker strings defining symbols to place
            at given locations.  If a single string is provided, that symbol
            will be placed at all locations.
        text_color : str
            Matplotlib color defining the color of the label text.
        ax : Axis
            Axis to plot on.  None will use the current axis.

        """
        ax = self._parse_ax(ax)
        if type(symbols) is str:
            symbols = [symbols] * len(labels)
        if len(labels) != len(locations):
            raise ValueError('length of labels and locations must match')
        if len(labels) != len(symbols):
            raise ValueError('length of labels and symbols must match')

        for loc, label, sym in zip(locations, labels, symbols):
            self.plot_label(label, loc, sym, text_color, ax)

    def plot_label(self, label, location, symbol='r+', text_color='k',
                   ax=None):
        """
        Plot a single symbol and label at a given location.

        Parameters
        ----------
        label : str
            Label text to place just above symbol.
        location : 2-tuples
            Tuple of latitude, longitude (in degrees) at which the symbol
            will be place.  The label is placed just above the symbol.
        symbol : str
            Matplotlib color+marker strings defining the symbol to place
            at the given location.
        text_color : str
            Matplotlib color defining the color of the label text.
        ax : Axis
            Axis to plot on.  None will use the current axis.

        """
        ax = self._parse_ax(ax)
        loc_x, loc_y = corner_to_point(self.loc, location)
        loc_x /= 1000.0
        loc_y /= 1000.0
        ax.plot([loc_x], [loc_y], symbol)
        ax.text(loc_x - 5.0, loc_y, label, color=text_color)

    def plot_cross_hair(self, size, npts=100, ax=None):
        """
        Plot a cross-hair on a ppi plot.

        Parameters
        ----------
        size : float
            Size of cross-hair in km.
        npts: int
            Number of points in the cross-hair, higher for better resolution.
        ax : Axis
            Axis to plot on.  None will use the current axis.

        """
        ax = self._parse_ax(ax)
        x = np.zeros(npts, dtype=np.float32)
        y = np.linspace(-size, size, npts)
        ax.plot(x, y, 'k-')  # verticle
        ax.plot(y, x, 'k-')  # horizontal

    def plot_colorbar(self, mappable=None, field=None, label=None, orient=None,
                      cax=None, ax=None, fig=None):
        """
        Plot a colorbar.

        Parameters
        ----------
        mappable : Image, ContourSet, etc.
            Image, ContourSet, etc to which the colorbar applied.  If None the
            last mappable object will be used.
        field : str
            Field to label colorbar with.
        label : str
            Colorbar label.  None will use a default value from the last field
            plotted.
        orient : str
            Colorbar orientation, either 'vertical' [default] or 'horizontal'.
        cax : Axis
            Axis onto which the colorbar will be drawn.  None is also valid.
        ax : Axes
            Axis onto which the colorbar will be drawn. None is also valid.
        fig : Figure
            Figure to place colorbar on.  None will use the current figure.

        """
        if fig is None:
            fig = plt.gcf()
        if mappable is None:
            mappable = self.plots[-1]
        if label is None:
            if field is None:
                field = self.plot_vars[-1]
            label = self._get_colorbar_label(field)
        # Find the axes locations to set colorbar
        box = make_axes_locatable(ax)

        if orient is None:
            orient = 'vertical'
        if orient == 'vertical':
            cax = box.append_axes("right", size="3%", pad=0.05)
        if orient == 'horizontal':
            cax = box.append_axes("bottom", size="6%", pad=0.50)
        cb = fig.colorbar(mappable, orientation=orient, ax=ax, cax=cax)
        cb.set_label(label)

    ##########################
    # Plot adjusting methods #
    ##########################

    def set_limits(self, xlim=None, ylim=None, ax=None):
        """
        Set the display limits.

        Parameters
        ----------
        xlim : tuple, optional
            2-Tuple containing y-axis limits in km. None uses default limits.
        ylim : tuple, optional
            2-Tuple containing x-axis limits in km. None uses default limits.
        ax : Axis
            Axis to adjust.  None will adjust the current axis.

        """
        if ax is None:
            ax = plt.gca()
        if ylim is not None:
            ax.set_ylim(ylim)
        if xlim is not None:
            ax.set_xlim(xlim)

    def label_xaxis_x(self, ax=None):
        """ Label the xaxis with the default label for x units. """
        ax = self._parse_ax(ax)
        ax.set_xlabel('Horizontal distance from ' + self.origin + ' (km)')

    def label_yaxis_y(self, ax=None):
        """ Label the yaxis with the default label for y units. """
        ax = self._parse_ax(ax)
        ax.set_ylabel('Horizontal distance from ' + self.origin + ' (km)')

    def label_yaxis_z(self, ax=None):
        """ Label the yaxis with the default label for y units. """
        ax = self._parse_ax(ax)
        ax.set_ylabel('Altitude (km)')

    def label_xaxis_r(self, ax=None):
        """ Label the xaxis with the default label for r units. """
        ax = self._parse_ax(ax)
        ax.set_xlabel('Distance from ' + self.origin + ' (km)')

    def label_xaxis_rays(self, ax=None):
        """ Label the yaxis with the default label for rays. """
        ax = self._parse_ax(ax)
        ax.set_xlabel('Ray number (unitless)')

    def label_yaxis_field(self, field, ax=None):
        """ Label the yaxis with the default label for a field units. """
        ax = self._parse_ax(ax)
        ax.set_ylabel(self._get_colorbar_label(field))

    def set_aspect_ratio(self, aratio=None, ax=None):
        """ Set the aspect ratio for plot area. """
        ax = self._parse_ax(ax)
        if aratio is None:
            ax.set_aspect(0.75)
        else:
            ax.set_aspect(aratio)

    def _set_title(self, field, title, ax):
        """ Set the figure title using a default title. """
        if title is None:
            ax.set_title(self.generate_title(field))
        else:
            ax.set_title(title)

    def _set_ray_title(self, field, ray, title, ax):
        """ Set the figure title for a ray plot using a default title. """
        if title is None:
            ax.set_title(self.generate_ray_title(field, ray))
        else:
            ax.set_title(title)

    def _label_axes_sweep(self, axis_labels, ax):
        """ Set the x and y axis labels for a PPI plot. """
        x_label, y_label = axis_labels
        if x_label is None:
            self.label_xaxis_x(ax)
        else:
            ax.set_xlabel(x_label)
        if y_label is None:
            self.label_yaxis_y(ax)
        else:
            ax.set_ylabel(y_label)

    def _label_axes_ppi(self, axis_labels, ax):
        """ Set the x and y axis labels for a PPI plot. """
        x_label, y_label = axis_labels
        if x_label is None:
            self.label_xaxis_x(ax)
        else:
            ax.set_xlabel(x_label)
        if y_label is None:
            self.label_yaxis_y(ax)
        else:
            ax.set_ylabel(y_label)

    def _label_axes_rhi(self, axis_labels, ax):
        """ Set the x and y axis labels for a RHI plot. """
        x_label, y_label = axis_labels
        if x_label is None:
            self.label_xaxis_r(ax)
        else:
            ax.set_xlabel(x_label)
        if y_label is None:
            self.label_yaxis_z(ax)
        else:
            ax.set_ylabel(y_label)

    def _label_axes_ray(self, axis_labels, field, ax):
        """ Set the x and y axis labels for a ray plot. """
        x_label, y_label = axis_labels
        if x_label is None:
            self.label_xaxis_r(ax)
        else:
            ax.set_xlabel(x_label)
        if y_label is None:
            self.label_yaxis_field(field, ax)
        else:
            ax.set_ylabel(y_label)

    ##########################
    # name generator methods #
    ##########################

    def generate_filename(self, field, sweep, ext='png'):
        """
        Generate a filename for a plot.

        Generated filename has form:
            radar_name_field_time.ext

        Parameters
        ----------
        field : str
            Field plotted.
        sweep : int
            Sweep plotted.
        ext : str
            Filename extension.

        Returns
        -------
        filename : str
            Filename suitable for saving a plot.

        """
        name_s = self.radar_name.replace(' ', '_')
        field_s = field.replace(' ', '_')
        time_s = self.time_begin.strftime('%Y%m%d%H%M%S')
        time_s = time_s.replace('-', ':')
        return '%s_%s_%s.%s' % (name_s, field_s, time_s, ext)

    def generate_title(self, field):
        """
        Generate a title for a plot.

        Parameters
        ----------
        field : str
            Field plotted.

        Returns
        -------
        title : str
            Plot title.

        """
        time_str = self.time_begin.isoformat() + 'Z'
        fixed_angle = self.fixed_angle
        l1 = "%s %.1f Deg. %s " % (self.radar_name, fixed_angle, time_str)
        field_name = self._generate_field_name(field)
        return l1 + '\n' + field_name

    def generate_ray_title(self, field, ray):
        """
        Generate a title for a ray plot.

        Parameters
        ----------
        field : str
            Field plotted.
        ray : int
            Ray plotted.

        Returns
        -------
        title : str
            Plot title.

        """
        time_str = self.time_begin.isoformat() + 'Z'
        l1 = "%s %s" % (self.radar_name, time_str)
        azim = self.azimuths[ray]
        elev = self.elevations[ray]
        l2 = "Ray: %i  Elevation: %.1f Azimuth: %.1f" % (ray, azim, elev)
        field_name = self._generate_field_name(field)
        return l1 + '\n' + l2 + '\n' + field_name

    def _generate_field_name(self, field):
        """ Return a nice field name for a particular field. """
        if 'standard_name' in self.fields[field]:
            field_name = self.fields[field]['standard_name']
        elif 'long_name' in self.fields[field]:
            field_name = self.fields[field]['long_name']
        else:
            field_name = str(field)
        field_name = field_name.replace('_', ' ')
        field_name = field_name[0].upper() + field_name[1:]
        return field_name

    def _generate_colorbar_label(self, standard_name, units):
        """ Generate and return a label for a colorbar. """
        return standard_name.replace('_', ' ') + ' (' + units + ')'

    ####################
    # Parseing methods #
    ####################

    def _parse_ax(self, ax):
        """ Parse and return ax parameter. """
        if ax is None:
            ax = plt.gca()
        return ax

    def _parse_ax_fig(self, ax, fig):
        """ Parse and return ax and fig parameters. """
        if ax is None:
            ax = plt.gca()
        if fig is None:
            fig = plt.gcf()
        return ax, fig

    def _parse_vmin_vmax(self, field, vmin, vmax):
        """ Parse and return vmin and vmax parameters. """
        field_dict = self.fields[field]
        if vmin is None:
            if 'valid_min' in field_dict:
                vmin = field_dict['valid_min']
            else:
                vmin = -6   # default value
        if vmax is None:
            if 'valid_max' in field_dict:
                vmax = field_dict['valid_max']
            else:
                vmax = 100
        return vmin, vmax

    ###############
    # Get methods #
    ###############

    def _get_data(self, field, mask_tuple):
        """ Retrieve and return data from a plot function. """
        data = self.fields[field]['data'][:]

        if mask_tuple is not None:  # mask data if mask_tuple provided
            mask_field, mask_value = mask_tuple
            mdata = self.fields[mask_field]['data'][:]
            data = np.ma.masked_where(mdata < mask_value, data)
        return data

    def _get_ray_data(self, field, ray, mask_tuple):
        """ Retrieve and return ray data from a plot function. """
        data = self.fields[field]['data'][ray]

        if mask_tuple is not None:
            mask_field, mask_value = mask_tuple
            mdata = self.fields[mask_field]['data'][ray]
            data = np.ma.masked_where(mdata < mask_value, data)
        return data

    def _get_x_z(self, field):
        """ Retrieve and return x and y coordinate in km. """
        return self.x/1000., self.z/1000.

    def _get_x_y(self, field):
        """ Retrieve and return x and y coordinate in km. """
        return self.x/1000.0, self.y/1000.0

    def _get_x_y_z(self, field):
        """ Retrieve and return x, y, and z coordinate in km. """
        x = self.x / 1000.0
        y = self.y / 1000.0
        z = self.z / 1000.0
        return x, y, z

    def _get_colorbar_label(self, field):
        """ Return a colorbar label for a given field. """
        last_field_dict = self.fields[field]
        if 'standard_name' in last_field_dict:
            standard_name = last_field_dict['standard_name']
        elif 'long_name' in last_field_dict:
            standard_name = last_field_dict['long_name']
        else:
            standard_name = field

        if 'units' in last_field_dict:
            units = last_field_dict['units']
        else:
            units = '?'
        return self._generate_colorbar_label(standard_name, units)



"""
pyart.graph.plot_mdv
====================

Routines for plotting radar data from MDV file.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    MdvDisplay

.. autosummary::
    :toctree: generated/

    _get_default_range

"""

import numpy as np

from .radardisplay import RadarDisplay


class MdvDisplay(RadarDisplay):
    """
    A display object for creating plots from data in a MdvFile objects.

    Parameters
    ----------
    mdvfile : MdvFile
        MdvFile object to use for creating plots.

    Attributes
    ----------
    plots : list
        List of plots created.
    plot_vars : list
        List of fields plotted, order matches plot list.
    cbs : list
        List of colorbars created.
    loc : (float, float)
        Latitude and Longitude of radar in degrees.
    origin : 'Radar'
       'Radar'
    time_begin : datetime
        Time of first radar scan.
    radar_name : str
        Radar name.
    mdvfile : MdvFile
        MdvFile object use to create plots.

    """

    def __init__(self, mdvfile):
        """ Initialize the object. """
        self.plots = []
        self.plot_vars = []
        self.cbs = []
        lat = mdvfile.radar_info['latitude_deg']
        lon = mdvfile.radar_info['longitude_deg']
        self.loc = [lat, lon]
        self.origin = 'radar'
        self.time_begin = mdvfile.times['time_begin']
        self.radar_name = mdvfile.master_header['data_set_source']
        self.mdvfile = mdvfile

    # public methods which are overridden.

    def generate_title(self, field, sweep):
        """
        Generate a title for a plot.

        Parameters
        ----------
        field : str
            Field plotted.
        sweep : int
            Sweep plotted.

        Returns
        -------
        title : str
            Plot title.

        """
        if self.mdvfile.scan_type == 'rhi':
            fixed_angle = self.mdvfile.az_deg[sweep]
        else:
            fixed_angle = self.mdvfile.el_deg[sweep]

        time_str = self.time_begin.isoformat() + 'Z'
        l1 = "%s %.1f Deg. %s " % (self.radar_name, fixed_angle, time_str)
        field_name = FANCY_NAMES[field]
        return l1 + '\n' + field_name

    # private methods which are overridden

    def _parse_vmin_vmax(self, field, vmin, vmax):
        if vmin is None:
            vmin = _get_default_range(self.mdvfile, field)[0]
        if vmax is None:
            vmax = _get_default_range(self.mdvfile, field)[1]
        return vmin, vmax

    def _get_data(self, field, sweep, mask_tuple, filter_transitions):
        """ Retrieve and return data from a plot function. """
        field_num = self.mdvfile.fields.index(field)
        data = self.mdvfile.read_a_field(field_num)[sweep]
        if mask_tuple is not None:
            mask_field, mask_value = mask_tuple
            mask_field_num = self.mdvfile.fields.index(mask_field)
            mdata = self.mdvfile.read_a_field(mask_field_num)[sweep]
            data = np.ma.masked_where(np.ma.less(mdata, mask_value), data)
        return data

    def _get_x_y(self, field, sweep, edges, filter_transitions):
        """ Retrieve and return x and y coordinate in km. """
        # TODO perform interpolating if edges is True
        x = self.mdvfile.carts['x'][sweep] / 1000.0  # x coords in km
        y = self.mdvfile.carts['y'][sweep] / 1000.0  # y coords in km
        return x, y

    def _get_x_y_z(self, field, sweep, edges, filter_transitions):
        """ Retrieve and return x, y, and z coordinate in km. """
        # TODO perform interpolating if edges is True
        x = self.mdvfile.carts['x'][sweep] / 1000.0  # x coords in km
        y = self.mdvfile.carts['y'][sweep] / 1000.0  # y coords in km
        z = self.mdvfile.carts['z'][sweep] / 1000.0  # y coords in km
        return x, y, z

    def _get_colorbar_label(self, field):
        """ Return a colobar label for a given field. """
        return UNIT_NAMES[field]


def _get_default_range(mdvfile, field):
    """ Return the default range for a field. """
    def_ranges = {
        'DBMHC': [-100, 0],
        'DBMVC': [-100, 0],
        'DBZ': [-16.0, 64.0],
        'DBZ_F': [-16.0, 64.0],
        'DBZVC': [-16.0, 64.0],
        'DBZVC_F': [-16.0, 64.0],
        'VEL': [-1.0*mdvfile.radar_info['unambig_vel_mps'],
                mdvfile.radar_info['unambig_vel_mps']],
        'VEL_F': [-1.0*mdvfile.radar_info['unambig_vel_mps'],
                  mdvfile.radar_info['unambig_vel_mps']],
        'WIDTH': [0.0, 10.0],
        'WIDTH_F': [0.0, 10.0],
        'ZDR': [-3, 6.0],
        'ZDR_F': [-3, 6.0],
        'RHOHV': [0.6, 1.0],
        'RHOHV_F': [0.6, 1.0],
        'PHIDP': [0, 180.0],
        'PHIDP_F': [-180.0, 180.0],
        'KDP': [-2, 6],
        'KDP_F': [-2, 6],
        'NCP': [0, 1],
        'NCP_F': [0, 1]}
    return def_ranges[field]


FANCY_NAMES = {
    'DBMHC': "Horizontal recieved power",
    'DBMVC': "Vertical recieved power",
    'DBZ': 'Horizontal equivalent reflectivity factor',
    'DBZ_F': 'Horizontal equivalent reflectivity factor',
    'DBZVC': 'Vertical equivalent reflectivity factor',
    'DBZVC_F': 'Vertical equivalent reflectivity factor',
    'VEL': "Radial velocity of scatterers (positive away)",
    'VEL_F': "Radial velocity of scatterers (positive away)",
    'WIDTH': "Spectral Width",
    'WIDTH_F': "Spectral Width",
    'ZDR': "Differential reflectivity",
    'ZDR_F': "Differential reflectivity",
    'RHOHV': 'Co-Polar correlation coefficient',
    'RHOHV_F': 'Co-Polar Correlation Coefficient',
    'PHIDP': "Differential propigation phase",
    'PHIDP_F': "Differential propigation phase",
    'KDP': "Specific differential phase",
    'KDP_F': "Specific differential phase",
    'NCP': "Normalized coherent power",
    'NCP_F': "Normalized coherent power"}


UNIT_NAMES = {
    'DBMHC': " H Rec Power (dBm)",
    'DBMVC': "V Rec Power (dBm)",
    'DBZ': "Hz Eq. Ref. Fac (dBz)",
    'DBZ_F': "Hz Eq. Ref. Fac (dBz)",
    'DBZVC': "V Eq. Ref. Fac (dBz)",
    'DBZVC_F': "V Eq. Ref. Fac (dBz)",
    'VEL': "Rad. Vel. (m/s, +away)",
    'VEL_F': "Rad. Vel. (m/s, +away)",
    'WIDTH': "Spec. Width (m/s)",
    'WIDTH_F': "Spec. Width (m/s)",
    'ZDR': "Dif Refl (dB)",
    'ZDR_F': "Dif Refl (dB)",
    'RHOHV': "Cor. Coef (frac)",
    'RHOHV_F': "Cor. Coef (frac)",
    'PHIDP': "Dif Phase (deg)",
    'PHIDP_F': "Dif Phase (deg)",
    'KDP': "Spec Dif Ph. (deg/km)",
    'KDP_F': "Spec Dif Ph. (deg/km)",
    'NCP': "Norm. Coh. Power (frac)",
    'NCP_F': "Norm. Coh. Power (frac)"}



""" Unit Tests for Py-ART's correct/dealias.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import datetime

import netCDF4
import pyart
import numpy as np
from numpy.testing import assert_allclose
from numpy.testing.decorators import skipif
from numpy.testing import assert_raises


@skipif(not pyart.io._RSL_AVAILABLE)
def test_find_time_in_interp_sounde():
    target = datetime.datetime(2011, 5, 10, 11, 30, 8)
    interp_sounde = netCDF4.Dataset(pyart.testing.INTERP_SOUNDE_FILE)
    t = pyart.correct.dealias.find_time_in_interp_sonde(interp_sounde, target)
    height, speed, direction = t

    assert height.shape == (316,)
    assert speed.shape == (316,)
    assert direction.shape == (316,)

    assert height.dtype == 'float32'
    assert speed.dtype == 'float32'
    assert direction.dtype == 'float32'

    assert round(height[100], 2) == 2.32
    assert round(speed[100], 2) == 15.54
    assert round(direction[100], 2) == 231.8


@skipif(not pyart.io._RSL_AVAILABLE)
def test_dealias_sounding():
    radar, dealias_vel = perform_dealias()
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    assert dealias_vel['data'][13, 46] is np.ma.masked
    assert dealias_vel['data'][13, 47] is np.ma.masked


@skipif(not pyart.io._RSL_AVAILABLE)
def test_dealias_sounding_keep_original():
    radar, dealias_vel = perform_dealias(True)
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    assert dealias_vel['data'][13, 46] == -7.5
    assert dealias_vel['data'][13, 47] == 8.5


def perform_dealias(keep_original=False):
    """ Perform velocity dealiasing on reference data. """
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be dealiased.
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    height = np.linspace(150, 250, 10).astype('float32')
    speed = np.ones((10), dtype='float32') * 0.5
    direction = np.ones((10), dtype='float32') * 5.
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sounding_heights=height, sounding_wind_speeds=speed,
        sounding_wind_direction=direction, keep_original=keep_original)
    return radar, dealias_vel


@skipif(not pyart.io._RSL_AVAILABLE)
def test_dealias_last_radar_and_sounding():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    height = np.linspace(150, 250, 10).astype('float32')
    speed = np.ones((10), dtype='float32') * 0.5
    direction = np.ones((10), dtype='float32') * 5.
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sounding_heights=height, sounding_wind_speeds=speed,
        sounding_wind_direction=direction, last_radar=last_radar,
        last_vel_field='velocity')
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    return


@skipif(not pyart.io._RSL_AVAILABLE)
def test_dealias_last_radar_and_sounding():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, last_radar=last_radar, last_vel_field='velocity')
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    return


@skipif(not pyart.io._RSL_AVAILABLE)
def test_error_raising():

    # ValueError when no sounding or last_radar provided
    radar = pyart.testing.make_velocity_aliased_radar()
    assert_raises(ValueError, pyart.correct.dealias_fourdd, radar)

    # ValueError when bad sounding data is provided
    height = np.linspace(150, 250, 0).astype('float32')
    speed = np.ones((0), dtype='float32') * 0.5
    direction = np.ones((0), dtype='float32') * 5.
    assert_raises(
        ValueError, pyart.correct.dealias_fourdd, radar,
        sounding_heights=height, sounding_wind_speeds=speed,
        sounding_wind_direction=direction)
    return


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields['dealiased_velocity'] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields['velocity']['data'][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields['dealiased_velocity']['data'][13])

    # create plot
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi('velocity', 0, ax=ax1, colorbar_flag=False,
                title='', vmin=-10, vmax=10)
    rd.plot_ppi('dealiased_velocity', 0, ax=ax2, colorbar_flag=False,
                title='', vmin=-10, vmax=20)
    fig.savefig('dealias_plot.png')



""" Mathematical, signal processing and numerical routines

TODO
----
Put more stuff in here

"""

import numpy as np


def rolling_window(a, window):
    """ create a rolling window object for application of functions
    eg: result=np.ma.std(array, 11), 1)"""
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1], )
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def texture(myradar, var):
    """Determine a texture field using an 11pt stdev
    texarray=texture(pyradarobj, field)
    """
    fld = myradar.fields[var]['data']
    print fld.shape
    tex = np.ma.zeros(fld.shape)
    for timestep in range(tex.shape[0]):
        ray = np.ma.std(rolling_window(fld[timestep, :], 11), 1)
        tex[timestep, 5:-5] = ray
        tex[timestep, 0:4] = np.ones(4) * ray[0]
        tex[timestep, -5:] = np.ones(5) * ray[-1]
    return tex



"""
pyart.retrieve.echo_class
=========================

steiner_conv_strat

.. autosummary::
    :toctree: generated/

    steiner_conv_strat

"""

import numpy as np

from ..config import get_fillvalue, get_field_name
from . import _echo_steiner


def steiner_conv_strat(grid, dx=None, dy=None, intense=42.0,
                       work_level=3000.0, peak_relation='default',
                       area_relation='medium', bkg_rad=11000.0,
                       use_intense=True, fill_value=None,
                       refl_field=None):
    """
    Partition reflectivity into convective-stratiform using the Steiner et
    al. (1995) algorithm.

    Parameters
    ----------
    grid : Grid
        Grid containing reflectivity field to partition.

    Other Parameters
    ----------------
    dx, dy : float
        The x- and y-dimension resolutions in meters, respectively.  If None
        the resolution is determined from the first two axes values.
    intense : float
        The intensity value in dBZ. Grid points with a reflectivity
        value greater or equal to the intensity are automatically
        flagged as convective. See reference for more information.
    work_level : float
        The working level (separation altitude) in meters. This is the height
        at which the partitioning will be done, and should minimize bright band
        contamination. See reference for more information.
    peak_relation : 'default' or 'sgp'
        The peakedness relation. See reference for more information.
    area_relation : 'small', 'medium', 'large', or 'sgp'
        The convective area relation. See reference for more information.
    bkg_rad : float
        The background radius in meters. See reference for more information.
    use_intense : bool
        True to use the intensity criteria.
    fill_value : float
         Missing value used to signify bad data points. A value of None
         will use the default fill value as defined in the Py-ART
         configuration file.
    refl_field : str
         Field in grid to use as the reflectivity during partitioning. None
         will use the default reflectivity field name from the Py-ART
         configuration file.

    Returns
    -------
    eclass : dict
        Steiner convective-stratiform classification dictionary.

    References
    ----------
    Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
    Characterization of Three-Dimensional Storm Structure from Operational
    Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.
    """

    # Get fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # Parse field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')

    # parse dx and dy
    if dx is None:
        dx = grid.axes['x_disp']['data'][1] - grid.axes['x_disp']['data'][0]
    if dy is None:
        dy = grid.axes['y_disp']['data'][1] - grid.axes['y_disp']['data'][0]

    # Get axes
    x = grid.axes['x_disp']['data']
    y = grid.axes['y_disp']['data']
    z = grid.axes['z_disp']['data']

    # Get reflectivity data
    ze = np.ma.copy(grid.fields[refl_field]['data'])
    ze = np.ma.filled(ze, fill_value).astype(np.float64)

    # Call Fortran routine
    eclass = _echo_steiner.classify(
        ze, x, y, z, dx=dx, dy=dy, bkg_rad=bkg_rad, work_level=work_level,
        intense=intense, peak_relation=peak_relation,
        area_relation=area_relation, use_intense=use_intense,
        fill_value=fill_value)

    return {'data': eclass.astype(np.int32),
            'standard_name': 'echo_classification',
            'long_name': 'Steiner echo classification',
            'valid_min': 0,
            'valid_max': 2,
            'comment_1': ('Convective-stratiform echo '
                         'classification based on '
                         'Steiner et al. (1995)'),
            'comment_2': ('0 = Undefined, 1 = Stratiform, '
                          '2 = Convective')}



""" Functions for converting date and time between various forms """

from netCDF4 import num2date


def datetime_from_radar(radar):
    """ Return a datetime for the first ray in a Radar. """
    return num2date(radar.time['data'][0], radar.time['units'])


def datetimes_from_radar(radar):
    """ Return an array of datetimes for the rays in a Radar. """
    return num2date(radar.time['data'][:], radar.time['units'])


def datetime_from_dataset(dataset):
    """ Return a datetime for the first time in a netCDF Dataset. """
    return num2date(dataset.variables['time'][0],
                    dataset.variables['time'].units)


def datetimes_from_dataset(dataset):
    """ Return an array of datetimes for the times in a netCDF Dataset. """
    return num2date(dataset.variables['time'][:],
                    dataset.variables['time'].units)


def datetime_from_grid(grid):
    """ Return a datetime for the volume start in a Grid. """
    return num2date(grid.axes['time_start']['data'][0],
                    grid.axes['time_start']['units'])



""" Unit Tests for Py-ART's correct/unwrap.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

from __future__ import print_function

import pyart
import numpy as np
from numpy.testing import assert_allclose
from numpy.testing import assert_raises

REF_DATA = [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
            12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
            2.5, 1.5, 0.5]


def test_dealias_unwrap_phase_ray():
    radar, dealias_vel = perform_dealias('ray')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_sweep():
    radar, dealias_vel = perform_dealias('sweep')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_volume():
    radar, dealias_vel = perform_dealias('volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_no_gatefilter():
    radar, dealias_vel = perform_dealias('sweep', gatefilter=False)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_explicit_gatefilter():
    radar = pyart.testing.make_velocity_aliased_radar()
    gf = pyart.correct.GateFilter(radar)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, gatefilter=gf)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_masked_field_sweep():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_masked_field_volume():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_masked_field_ray():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='ray')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_rhi_sweep():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_rhi_volume():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_raises():

    # invalid unwrap_unit
    radar = pyart.testing.make_velocity_aliased_radar()
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  unwrap_unit='fuzz')

    # no explicit nyquist
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.instrument_parameters = None
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # non-sequential
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.azimuth['data'][10] = 190.
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # non-aligned sweeps
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields['velocity'] = {'data': np.zeros((20, 1))}
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  nyquist_vel=10, unwrap_unit='volume')

    # non-cubic radar
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields['velocity'] = {'data': np.zeros((20, 1))}
    radar.azimuth['data'][-10:] = range(10)
    radar.sweep_end_ray_index['data'][-1] = 18
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  nyquist_vel=10, unwrap_unit='volume')

    # invalid scan type
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.scan_type = 'fuzz'
    assert_raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)


def perform_dealias(unwrap_unit='sweep', **kwargs):
    """ Perform velocity dealiasing on reference data. """
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be not be dealiased.
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit=unwrap_unit, **kwargs)
    return radar, dealias_vel


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields['dealiased_velocity'] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields['velocity']['data'][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields['dealiased_velocity']['data'][13])

    # create plot
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi('velocity', 0, ax=ax1, colorbar_flag=False,
                title='', vmin=-10, vmax=10)
    rd.plot_ppi('dealiased_velocity', 0, ax=ax2, colorbar_flag=False,
                title='', vmin=-10, vmax=20)
    fig.savefig('dealias_plot.png')



""" Unit Tests for Py-ART's graph/gridmapdisplay.py module. """
# execute this script to create figure_gridmapdisplay_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import pyart
from numpy.testing import assert_raises
from numpy.testing.decorators import skipif

RESOLUTION = 'c'    # crude resolution to speed up tests


@skipif('GridMapDisplay' not in dir(pyart.graph))
def test_gridmapdisplay_simple(outfile=None):
    # test basic GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_basemap(ax=ax, resolution=RESOLUTION)
    display.plot_grid('reflectivity', vmin=-5., vmax=35)
    if outfile:
        fig.savefig(outfile)


@skipif('GridMapDisplay' not in dir(pyart.graph))
def test_gridmapdisplay_fancy(outfile=None):
    # test a bunch of GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()

    ax = fig.add_subplot(321)
    display.plot_basemap(ax=ax, resolution=RESOLUTION)
    display.plot_grid('reflectivity', vmin=-5., vmax=35)
    display.plot_crosshairs(line_style='b--')

    ax = fig.add_subplot(322)
    display.plot_colorbar()

    ax = fig.add_subplot(323)
    display.plot_latitude_slice('reflectivity')  # no vmin/vmax

    ax = fig.add_subplot(325)
    grid.fields['reflectivity']['valid_min'] = 0.
    grid.fields['reflectivity']['valid_max'] = 30.
    display.plot_longitude_slice('reflectivity')

    ax = fig.add_subplot(326)
    del display.grid.fields['reflectivity']['long_name']
    display.plot_colorbar()

    if outfile:
        fig.savefig(outfile)


@skipif('GridMapDisplay' not in dir(pyart.graph))
def test_error_raising():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    # no basemap
    assert_raises(ValueError, display.plot_grid, 'reflectivity')

    # no mappable
    assert_raises(ValueError, display.plot_colorbar)

    # no field
    display.mappables.append(None)  # mock the mappable
    assert_raises(ValueError, display.plot_colorbar)


if __name__ == "__main__":
    test_gridmapdisplay_simple('figure_gridmapdisplay_simple.png')
    test_gridmapdisplay_fancy('figure_gridmapdisplay_fancy.png')



""" Unit Tests for Py-ART's graph/plot_rsl.py module. """
# execute this script to create figure_plot_rsl_*.png files.
# To compare results:
# display figure_plot_rsl_ppi.png & display figure_plot_rsl_ppi_radar.png &
# display figure_plot_rsl_rhi.png & display figure_plot_rsl_rhi_radar.png &

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
from numpy.testing.decorators import skipif
import pyart


@skipif(not pyart.io._RSL_AVAILABLE)
def test_plot_rsl_rhi(outfile=None):
    rslfile = pyart.io._rsl_interface.RslFile(pyart.testing.SIGMET_RHI_FILE)
    display = pyart.graph.RslDisplay(rslfile)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('DZ', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


@skipif(not pyart.io._RSL_AVAILABLE)
def test_plot_rsl_ppi(outfile=None):
    rslfile = pyart.io._rsl_interface.RslFile(pyart.testing.SIGMET_PPI_FILE)
    display = pyart.graph.RslDisplay(rslfile)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('DZ', 0, ax=ax)
    if outfile:
        fig.savefig(outfile)


# version of above which use the RadarDisplay to return similar images.


@skipif(not pyart.io._RSL_AVAILABLE)
def test_plot_rsl_radar_rhi(outfile=None):
    radar = pyart.io.read_rsl(pyart.testing.SIGMET_RHI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_rhi('reflectivity', 0,
                     vmin=-16, vmax=64, ax=ax)
    if outfile:
        fig.savefig(outfile)


@skipif(not pyart.io._RSL_AVAILABLE)
def test_plot_rsl_radar_ppi(outfile=None):
    radar = pyart.io.read_rsl(pyart.testing.SIGMET_PPI_FILE)
    display = pyart.graph.RadarDisplay(radar)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi('reflectivity', 0,
                     vmin=-16, vmax=64, ax=ax)
    if outfile:
        fig.savefig(outfile)


if __name__ == "__main__":
    test_plot_rsl_rhi('figure_plot_rsl_rhi.png')
    test_plot_rsl_ppi('figure_plot_rsl_ppi.png')
    test_plot_rsl_radar_rhi('figure_plot_rsl_rhi_radar.png')
    test_plot_rsl_radar_ppi('figure_plot_rsl_ppi_radar.png')



""" Unit Tests for Py-ART's correct/region_dealias.py module. """

# python test_
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

from __future__ import print_function

import pyart
import numpy as np
from numpy.testing import assert_allclose

REF_DATA = [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
            12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
            2.5, 1.5, 0.5]


def test_dealias_region_based():
    radar, dealias_vel = perform_dealias()
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_region_based_masked_wrap_around():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][0, 25] = np.ma.masked
    radar.fields['velocity']['data'][-1, 30] = np.ma.masked
    radar.fields['velocity']['data'][180, -1] = np.ma.masked
    radar.fields['velocity']['data'][180, 0] = np.ma.masked
    dealias_vel = pyart.correct.dealias_region_based(
        radar, rays_wrap_around=True)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][0, 25])
    assert not np.ma.is_masked(dealias_vel['data'][0, 24])
    assert np.ma.is_masked(dealias_vel['data'][-1, 30])
    assert not np.ma.is_masked(dealias_vel['data'][-1, 29])
    assert np.ma.is_masked(dealias_vel['data'][180, -1])
    assert not np.ma.is_masked(dealias_vel['data'][180, -2])
    assert np.ma.is_masked(dealias_vel['data'][180, 0])
    assert not np.ma.is_masked(dealias_vel['data'][180, 1])


def test_dealias_region_based_no_wrap_around():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][0, 25] = np.ma.masked
    radar.fields['velocity']['data'][0, 25] = np.ma.masked
    radar.fields['velocity']['data'][-1, 30] = np.ma.masked
    dealias_vel = pyart.correct.dealias_region_based(
        radar, rays_wrap_around=False)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][0, 25])
    assert not np.ma.is_masked(dealias_vel['data'][0, 24])
    assert np.ma.is_masked(dealias_vel['data'][-1, 30])
    assert not np.ma.is_masked(dealias_vel['data'][-1, 29])


def perform_dealias(**kwargs):
    """ Perform velocity dealiasing on reference data. """
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be not be dealiased.
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_region_based(
        radar, **kwargs)
    return radar, dealias_vel


def main():

    radar, dealias_vel = perform_dealias()
    radar.fields['dealiased_velocity'] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields['velocity']['data'][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields['dealiased_velocity']['data'][13])

    # create plot
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi('velocity', 0, ax=ax1, colorbar_flag=False,
                title='', vmin=-10, vmax=10)
    rd.plot_ppi('dealiased_velocity', 0, ax=ax2, colorbar_flag=False,
                title='', vmin=-10, vmax=20)
    plt.show()
    fig.savefig('dealias_plot.png')


if __name__ == "__main__":
    main()



""" Unit Tests for Py-ART's map/grid_mapper.py module. """

import numpy as np
from numpy.testing import assert_array_equal, assert_raises

import pyart

EXPECTED_CENTER_SLICE = [40, 30, 20, 10, 0, 0, 10, 20, 30, 40]

COMMON_MAP_TO_GRID_ARGS = {
    'grid_shape': (3, 9, 10),
    'grid_limits': ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
    'fields': None,
    'roi_func': lambda x, y, z: 30, }


def test_map_to_grid_default():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar,),
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_cressman():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,), (3, 9, 10), ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func='constant', constant_roi=30., weighting_function='CRESSMAN')
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_constant_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,), (3, 9, 10), ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func='constant', constant_roi=30.)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_dist_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,), (3, 9, 10), ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func='dist', z_factor=0, xy_factor=0, min_radius=30.)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_dist_beam_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        grid_shape=(3, 9, 10),
        grid_limits=((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        fields=['reflectivity'],
        min_radius=30, bsp=0., h_factor=0., max_refl=100)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_default_two_radars():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar, radar),
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_masked_refl_field():
    radar = pyart.testing.make_target_radar()

    # mask the last gate of the first ray
    fdata = radar.fields['reflectivity']['data']
    fdata = np.ma.masked_invalid(fdata)
    fdata.mask[0, -1] = True
    radar.fields['reflectivity']['data'] = fdata

    grids = pyart.map.map_to_grid((radar,),
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_no_copy():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar,), copy_field_data=False,
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_no_copy_two_radars():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar, radar), copy_field_data=False,
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_balltree():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar,), algorithm='ball_tree',
                                  **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids['reflectivity'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)


def test_map_to_grid_tiny_grid():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        grid_shape=(1, 1, 1),
        grid_limits=((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        fields=['reflectivity'])
    assert grids['reflectivity'].shape == (1, 1, 1)
    assert int(grids['reflectivity'][0]) == 40


def test_map_to_grid_errors():
    assert_raises(ValueError, pyart.map.map_to_grid, (None, ), (1, 1, 1),
                  ((-1, 1), (-1, 1), (-1, 1)),
                  weighting_function='foo')
    assert_raises(ValueError, pyart.map.map_to_grid, (None, ), (1, 1, 1),
                  ((-1, 1), (-1, 1), (-1, 1)),
                  algorithm='foo')
    radar = pyart.testing.make_target_radar()
    assert_raises(ValueError, pyart.map.map_to_grid, (radar, ), (1, 1, 1),
                  ((-1, 1), (-1, 1), (-1, 1)),
                  roi_func='foo')


def test_grid_from_radars():
    radar = pyart.testing.make_target_radar()
    grid = pyart.map.grid_from_radars((radar,), **COMMON_MAP_TO_GRID_ARGS)

    # check field data
    center_slice = grid.fields['reflectivity']['data'][1, 4, :]
    assert_array_equal(np.round(center_slice), EXPECTED_CENTER_SLICE)

    # check other Grid object attributes
    assert 'ROI' in grid.fields
    assert np.all(grid.fields['ROI']['data'] == 30.)
    assert_array_equal(grid.axes['x_disp']['data'],
                       np.linspace(-900, 900, 10))
    assert_array_equal(grid.axes['y_disp']['data'],
                       np.linspace(-900, 900, 9).astype('float64'))
    assert_array_equal(grid.axes['z_disp']['data'],
                       np.linspace(-400, 400, 3).astype('float64'))


def test_grid_from_radars_grid_origin():
    radar = pyart.testing.make_target_radar()
    radar.metadata.pop('instrument_name')
    grid = pyart.map.grid_from_radars((radar,), grid_origin=(36.4, -97.6),
                                      **COMMON_MAP_TO_GRID_ARGS)
    print round(grid.axes['lat']['data'][0], 2)
    print round(grid.axes['lon']['data'][0], 2)
    assert round(grid.axes['lat']['data'][0], 2) == 36.4
    assert round(grid.axes['lon']['data'][0], 2) == -97.6


def test_example_roi_funcs():
    assert pyart.map.example_roi_func_constant(0, 0, 0) == 500.
    assert pyart.map.example_roi_func_dist(0, 0, 0) == 500.
    assert pyart.map.example_roi_func_dist_beam(0, 0, 0) == 500.



"""
=============================
Graphing (:mod:`pyart.graph`)
=============================

.. currentmodule:: pyart.graph

.. autosummary::
    :toctree: generated/

    RadarDisplay
    RadarMapDisplay
    GridMapDisplay
    MdvDisplay
    RslDisplay
    CFRadialDisplay
    RadarDisplay_Airborne


"""

from .radardisplay import RadarDisplay
from .plot_mdv import MdvDisplay
from .plot_cfradial import CFRadialDisplay
from . import cm
from .radardisplay_airborne import RadarDisplay_Airborne

try:
    from .gridmapdisplay import GridMapDisplay
except ImportError:
    import warnings
    warnings.warn('No grid plotting support, requires basemap.')

try:
    from .radarmapdisplay import RadarMapDisplay
except ImportError:
    import warnings
    warnings.warn('No grid plotting support, requires basemap.')

try:
    from .plot_rsl import RslDisplay
except ImportError:
    pass

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.graph.plot_rsl
====================

Routines for plotting radar data from files readable by RSL.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    RslDisplay

"""

# Nothing from this module is imported into pyart.graph is RSL is not
# installed.
import numpy as np

from .common import radar_coords_to_cart, dms_to_d

from ..io import rsl
from .radardisplay import RadarDisplay


# XXX currenly broken
class RslDisplay(RadarDisplay):
    """
    A display object for creating plots from data in a RslFile object.

    Parameters
    ----------
    rslfile : RslFile
        RslFile object to use for creating plots.

    Attributes
    ----------
    plots : list
        List of plots created.
    plot_vars : list
        List of fields plotted, order matches plot list.
    cbs : list
        List of colorbars created.
    loc : (float, float)
        Latitude and Longitude of radar in degrees.
    origin : 'Radar'
       'Radar'
    time_begin : datetime
        Time of first radar scan.
    radar_name : str
        Radar name.
    rslfile : RslFile
        RSL radar object use to create plots.

    """

    def __init__(self, rslfile):
        """ Initialize the object. """
        self.plots = []
        self.plot_vars = []
        self.cbs = []

        # radar location in lat, lon
        latd = rslfile.latd
        latm = rslfile.latm
        lats = rslfile.lats

        lond = rslfile.lond
        lonm = rslfile.lonm
        lons = rslfile.lons

        lat = dms_to_d((latd, latm, lats))
        lon = dms_to_d((lond, lonm, lons))
        self.loc = [lat, lon]
        self.origin = 'radar'
        dt = rslfile.get_volume(0).get_sweep(0).get_ray(0).get_datetime()
        self.time_begin = dt

        self.radar_name = rslfile.get_radar_header()['radar_name']
        self.rslfile = rslfile

    # public methods which are overridden.

    def generate_title(self, field, sweep):
        """
        Generate a title for a plot.

        Parameters
        ----------
        field : str
            Field plotted.
        sweep : int
            Sweep plotted.

        Returns
        -------
        title : str
            Plot title.

        """
        volume_num = rsl.RSLNAME2VOLUMENUM[field]
        sweep = self.rslfile.get_volume(volume_num).get_sweep(sweep)
        fixed_angle = sweep.azimuth
        if fixed_angle == -999.0:   # ppi scan
            fixed_angle = sweep.elev

        time_str = self.time_begin.isoformat() + 'Z'
        l1 = "%s %.1f Deg. %s " % (self.radar_name, fixed_angle, time_str)
        return l1 + '\n' + field

    # private methods which are overridden

    def _parse_vmin_vmax(self, field, vmin, vmax):
        if vmin is None:
            if field in DEFAULT_VMIN_VMAX:
                vmin = DEFAULT_VMIN_VMAX[field][0]
            else:
                vmin = -16.0
        if vmax is None:
            if field in DEFAULT_VMIN_VMAX:
                vmax = DEFAULT_VMIN_VMAX[field][1]
            else:
                vmax = 64.0
        return vmin, vmax

    def _get_data(self, field, sweep, mask_tuple, filter_transitions):
        """ Retrieve and return data from a plot function. """
        volume_num = rsl.RSLNAME2VOLUMENUM[field]
        sweep = self.rslfile.get_volume(volume_num).get_sweep(sweep)
        data = sweep.get_data()

        if mask_tuple is not None:
            mask_field, mask_value = mask_tuple
            mask_num = rsl.RSLNAME2VOLUMENUM[mask_field]
            sweep = self.rslfile.get_volume(mask_num).get_sweep(sweep)
            mdata = sweep.get_data()
            data = np.ma.masked_where(mdata < mask_value, data)
        return data

    def _get_x_y(self, field, sweep, edges, filter_transitions):
        """ Retrieve and return x and y coordinate in km. """
        # TODO perform interpolating if edges is True
        x, y, z = self._get_x_y_z(field, sweep, edges, filter_transitions)
        return x, y

    def _get_x_y_z(self, field, sweep, edges, filter_transitions):
        """ Retrieve and return x, y, and z coordinate in km. """
        # TODO perform interpolating if edges is True
        volume_num = rsl.RSLNAME2VOLUMENUM[field]
        sweep = self.rslfile.get_volume(volume_num).get_sweep(sweep)
        ray = sweep.get_ray(0)
        ranges = ray.range_bin1 + ray.gate_size * np.arange(ray.nbins)
        ranges = ranges / 1000.0
        elevs = [sweep.get_ray(i).elev for i in xrange(sweep.nrays)]
        azimuths = [sweep.get_ray(i).azimuth for i in xrange(sweep.nrays)]
        rg, ele = np.meshgrid(ranges, elevs)
        rg, azg = np.meshgrid(ranges, azimuths)
        x, y, z = radar_coords_to_cart(rg, azg, ele)
        return x / 1000.0, y / 1000.0, z / 1000.0

    def _get_colorbar_label(self, field):
        """ Return a colobar label for a given field. """
        if field in DEFAULT_COLORBAR_LABELS:
            return DEFAULT_COLORBAR_LABELS[field]
        else:
            return "Unknown units"


DEFAULT_VMIN_VMAX = {
    'CZ': [-16., 64.],
    'DZ': [-16., 64.],
    'ZT': [-16.0, 64.0],
    'VE': [-40.0, 40.0],
    'VR': [-16, 16],
    'PH': [300, 350],
    'DR': [-4, 3]}

DEFAULT_COLORBAR_LABELS = {
    'CZ': 'Eq refl fact (dBz)',
    'DZ': 'Eq refl fact (dBz)',
    'ZT': 'Eq refl fact (dBz)',
    'PH': 'Phidp deg',
    'VE': 'Unfolded VR (m/s)',
    'DR': 'ZDR dB',
    'VR': 'Radial Velocity (+away) m/s'}



""" Unit Tests for Py-ART's core/grid.py module. """

import tempfile
import os

import numpy as np

import pyart

COMMON_MAP_TO_GRID_ARGS = {
    'grid_shape': (3, 9, 10),
    'grid_limits': ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
    'fields': ['reflectivity'],
    'roi_func': lambda x, y, z: 30, }


def test_grid_from_radars():
    radar = pyart.testing.make_target_radar()
    grid = pyart.map.grid_from_radars((radar,), **COMMON_MAP_TO_GRID_ARGS)

    tmpfile = tempfile.mkstemp(suffix='.nc', dir='.')[1]
    grid.write(tmpfile)
    grid2 = pyart.io.read_grid(tmpfile)

    # check metadata
    for k, v in grid.metadata.iteritems():
        print "Checking key:", k, "should have value:", v
        print grid2.metadata
        assert grid2.metadata[k] == v

    # check axes
    for axes_key in grid.axes.keys():
        for k, v in grid.axes[axes_key].iteritems():
            print "Checking axes_key:", axes_key, "key:", k
            if k == 'data':
                assert np.all(grid.axes[axes_key][k] == v)
            else:
                assert grid2.axes[axes_key][k] == v

    # check fields
    for field in grid.fields.keys():
        for k, v in grid.fields[field].iteritems():
            print "Checking field:", field, "key:", k
            if k == 'data':
                assert np.all(grid.fields[field][k] == v)
            else:
                assert grid2.fields[field][k] == v

    os.remove(tmpfile)
